#!/usr/bin/python
import logging

from mdg.tools.validate import validate_package
from mdg.generate.render import output_model, output_test_cases
from mdg.parse import parse, ParseError
from mdg.config import settings


logger = logging.getLogger('mdg')


def generate():
    """ Loads XMI file from settings as an etree
        Calls XMI parser to turn model and tests into python native (see UML metamodel)
        Calls output functions to render for model and tests
    """
    logger.debug("generate begin")
    try:
        model_package, test_cases = parse()

        # Validate the parsed model package
        errors = validate_package(model_package)
        if len(errors) > 0:
            print("Validation Errors:")
            for error in errors:
                print("    {}".format(error))

        # TODO: Validate the test cases

        # Generate files from the native python UML
        output_model(model_package)
        if 'test_package' in settings.keys():
            output_test_cases(test_cases)

    except ParseError as e:
        logger.error(e)
    logger.debug("generate end")



import json

from mdg.generate.confluence.util import basic_auth, http_put, http_get_json


def update(auth, content_id, path):
    url = "https://jiragenesisenergy.atlassian.net/wiki/rest/api/content/{}".format(content_id)
    content, session = http_get_json(url, auth)
    new_version = int(content['version']['number']) + 1

    attachments, session = http_get_json("https://jiragenesisenergy.atlassian.net/wiki/rest/api/content/{}/child/attachment".format(content_id), auth, session)

    with open(path) as f:
        body = f.read()

    data = {
        "id": "{}".format(content_id),
        "type": "page",
        "title": "Models - Retail Test",
        "space": {
            "key":
            "DA"},
        "body": {
            "storage": {
                "value": "{}".format(body),
                "representation": "storage"}},
        "version": {
            "number": new_version
        }
    }
    http_put(url, auth, json.dumps(data), None)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('user', help='confluence username')
    parser.add_argument('token', help='confluence API token')
    parser.add_argument('page', help='confluence page id')
    parser.add_argument('path', help='path to documentation content')
    args = parser.parse_args()

    auth = basic_auth(args.user, args.token)
    update(auth, args.page, args.path)



import unittest

from mdg.uml import UMLAttribute, UMLInstance, UMLPackage
from mdg.generate.render import serialize_instance


class TestInstanceOutput(unittest.TestCase):
    def setUp(self):
        self.package = UMLPackage("id", "name")
        self.instance = UMLInstance(self.package, "test1", 1)

        attr = UMLAttribute(self.package, 'Test Att', 1)
        attr.value = 'testval'
        self.instance.attributes.append(attr)

        attr = UMLAttribute(self.package, 'Test Att 2', 2)
        attr.value = 10
        self.instance.attributes.append(attr)

        self.package.instances.append(self.instance)

    def test_object_serialise(self):
        data = serialize_instance(self.instance)
        self.assertEqual(data['test_att'], 'testval')

        self.assertEqual(data['test_att_2'], 10)



import os
import yaml
from typing import Dict, List
import logging


logger = logging.getLogger(__name__)

model_templates: List[Dict] = []

defaults: Dict = {
    "generation_type": "default",
    "root_package": "default",
    "model_templates": model_templates,
    "case_package": "CamelCase",
    "case_class": "CamelCase",
    "case_attribute": "snake_case",
    "parser": None,
    "use_alias": True,
}


def load():
    config_filename = os.environ.get('PYMDG_SETTINGS_MODULE', "")
    try:
        with open(config_filename, 'r') as config_file:
            loaded_settings = yaml.load(config_file.read(), Loader=yaml.SafeLoader)
        settings = {**defaults, **loaded_settings}
        logger.info("Config file loaded: " + config_filename)
        return settings
    except TypeError as e:
        logger.warning("Config file {} could not be parsed. Using default settings. Reason: {}".format(config_filename, e.message))
    except FileNotFoundError:
        logger.warning("WARN: Config file {} not found. Using default settings".format(config_filename))
    return defaults


# If no recipe has been set before now, assume test recipe
if os.environ.get('PYMDG_SETTINGS_MODULE', None) is None:
    recipe_path = './sample_recipes' + "/config-test.yaml"
    os.environ.setdefault("PYMDG_SETTINGS_MODULE", recipe_path)

settings = load()




from sqlalchemy import (
    CHAR,
    Column,
    # Date,
    DateTime,
    # Float,
    Index,
    Integer,
    # LargeBinary,
    # SmallInteger,
    String,
    # Table,
    Text,
    text,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.decl_api import declarative_base as _declarative_base

Base: _declarative_base = declarative_base()
metadata = Base.metadata


class TObject(Base):
    __tablename__ = 't_object'
    __table_args__ = (
        Index('ix_object_eaguidclassifier', 'ea_guid', 'classifier'),
        Index('ix_object_pkgidpdata1class', 'package_id', 'pdata1', 'classifier')
    )

    object_id = Column(Integer, primary_key=True, server_default=text("nextval(('object_id_seq'::text)::regclass)"))
    object_type = Column(String(255), index=True)
    diagram_id = Column(Integer, server_default=text("0"))
    name = Column(String(255), index=True)
    alias = Column(String(255))
    author = Column(String(255))
    version = Column(String(50), server_default=text("'1.0'::character varying"))
    note = Column(Text)
    package_id = Column(Integer, index=True, server_default=text("0"))
    stereotype = Column(String(255))
    ntype = Column(Integer, index=True, server_default=text("0"))
    complexity = Column(String(50), server_default=text("'2'::character varying"))
    effort = Column(Integer, server_default=text("0"))
    style = Column(String(255))
    backcolor = Column(Integer, server_default=text("0"))
    borderstyle = Column(Integer, server_default=text("0"))
    borderwidth = Column(Integer, server_default=text("0"))
    fontcolor = Column(Integer, server_default=text("0"))
    bordercolor = Column(Integer, server_default=text("0"))
    createddate = Column(DateTime, server_default=text("now()"))
    modifieddate = Column(DateTime, server_default=text("now()"))
    status = Column(String(50))
    abstract = Column(CHAR(1))
    tagged = Column(Integer, server_default=text("0"))
    pdata1 = Column(String(255), index=True)
    pdata2 = Column(Text, index=True)
    pdata3 = Column(Text, index=True)
    pdata4 = Column(Text, index=True)
    pdata5 = Column(String(255), index=True)
    concurrency = Column(String(50))
    visibility = Column(String(50))
    persistence = Column(String(50))
    cardinality = Column(String(50))
    gentype = Column(String(50))
    genfile = Column(String(255))
    header1 = Column(Text)
    header2 = Column(Text)
    phase = Column(String(50))
    scope = Column(String(25))
    genoption = Column(Text)
    genlinks = Column(Text, index=True)
    classifier = Column(Integer, index=True)
    ea_guid = Column(String(40), unique=True)
    parentid = Column(Integer, index=True)
    runstate = Column(Text)
    classifier_guid = Column(String(40), index=True)
    tpos = Column(Integer)
    isroot = Column(Integer, server_default=text("0"))
    isleaf = Column(Integer, server_default=text("0"))
    isspec = Column(Integer, server_default=text("0"))
    isactive = Column(Integer, server_default=text("0"))
    stateflags = Column(String(255))
    packageflags = Column(String(255))
    multiplicity = Column(String(50))
    styleex = Column(Text)
    actionflags = Column(String(255))
    eventflags = Column(String(255), index=True)


class TPackage(Base):
    __tablename__ = 't_package'

    package_id = Column(Integer, primary_key=True, server_default=text("nextval(('package_id_seq'::text)::regclass)"))
    name = Column(String(255), index=True)
    parent_id = Column(Integer, index=True, server_default=text("0"))
    createddate = Column(DateTime, server_default=text("now()"))
    modifieddate = Column(DateTime, server_default=text("now()"))
    notes = Column(Text)
    ea_guid = Column(String(40), unique=True)
    xmlpath = Column(String(255))
    iscontrolled = Column(Integer, server_default=text("0"))
    lastloaddate = Column(DateTime)
    lastsavedate = Column(DateTime)
    version = Column(String(50))
    protected = Column(Integer, server_default=text("0"))
    pkgowner = Column(String(255))
    umlversion = Column(String(50))
    usedtd = Column(Integer, server_default=text("0"))
    logxml = Column(Integer, server_default=text("0"))
    codepath = Column(String(255))
    namespace = Column(String(50))
    tpos = Column(Integer)
    packageflags = Column(String(255))
    batchsave = Column(Integer)
    batchload = Column(Integer)


class TXref(Base):
    __tablename__ = 't_xref'
    __table_args__ = (
        Index('ix_xref_nametype', 'name', 'type'),
    )

    xrefid = Column(String(255), primary_key=True)
    name = Column(String(255), index=True)
    type = Column(String(255), index=True)
    visibility = Column(String(255))
    namespace = Column(String(255))
    requirement = Column(String(255))
    Constraint = Column(String(255))
    behavior = Column(String(255))
    partition = Column(String(255))
    description = Column(Text)
    client = Column(String(255), index=True)
    supplier = Column(String(255), index=True)
    link = Column(String(255))


class TAttribute(Base):
    __tablename__ = 't_attribute'

    object_id = Column(Integer, index=True, server_default=text("0"))
    name = Column(String(255), index=True)
    scope = Column(String(50))
    stereotype = Column(String(50))
    containment = Column(String(50))
    isstatic = Column(Integer, server_default=text("0"))
    iscollection = Column(Integer, server_default=text("0"))
    isordered = Column(Integer, server_default=text("0"))
    allowduplicates = Column(Integer, server_default=text("0"))
    lowerbound = Column(String(50))
    upperbound = Column(String(50))
    container = Column(String(50))
    notes = Column(Text)
    derived = Column(CHAR(1))
    id = Column(Integer, primary_key=True, server_default=text("nextval(('id_seq'::text)::regclass)"))
    pos = Column(Integer)
    genoption = Column(Text)
    length = Column(Integer)
    precision = Column(Integer)
    scale = Column(Integer)
    const = Column(Integer)
    style = Column(String(255))
    classifier = Column(String(50), index=True)
    Default = Column(Text)
    type = Column(String(255), index=True)
    ea_guid = Column(String(50), unique=True)
    styleex = Column(Text)


class TConnector(Base):
    __tablename__ = 't_connector'
    __table_args__ = (
        Index('ix_connector_startobjidconnid', 'start_object_id', 'connector_id'),
        Index('ix_connector_endobjidconnid', 'end_object_id', 'connector_id')
    )

    connector_id = Column(Integer, primary_key=True, server_default=text("nextval(('connector_id_seq'::text)::regclass)"))
    name = Column(String(255))
    direction = Column(String(50))
    notes = Column(Text)
    connector_type = Column(String(50), index=True)
    subtype = Column(String(50), index=True)
    sourcecard = Column(String(50))
    sourceaccess = Column(String(50))
    sourceelement = Column(String(50))
    destcard = Column(String(50))
    destaccess = Column(String(50))
    destelement = Column(String(50))
    sourcerole = Column(String(255))
    sourceroletype = Column(String(50))
    sourcerolenote = Column(Text)
    sourcecontainment = Column(String(50))
    sourceisaggregate = Column(Integer, server_default=text("0"))
    sourceisordered = Column(Integer, server_default=text("0"))
    sourcequalifier = Column(String(50))
    destrole = Column(String(255))
    destroletype = Column(String(50))
    destrolenote = Column(Text)
    destcontainment = Column(String(50))
    destisaggregate = Column(Integer, server_default=text("0"))
    destisordered = Column(Integer, server_default=text("0"))
    destqualifier = Column(String(50))
    start_object_id = Column(Integer, index=True, server_default=text("0"))
    end_object_id = Column(Integer, index=True, server_default=text("0"))
    top_start_label = Column(String(50))
    top_mid_label = Column(String(50))
    top_end_label = Column(String(50))
    btm_start_label = Column(String(50))
    btm_mid_label = Column(String(50))
    btm_end_label = Column(String(50))
    start_edge = Column(Integer, server_default=text("0"))
    end_edge = Column(Integer, server_default=text("0"))
    ptstartx = Column(Integer, server_default=text("0"))
    ptstarty = Column(Integer, server_default=text("0"))
    ptendx = Column(Integer, server_default=text("0"))
    ptendy = Column(Integer, server_default=text("0"))
    seqno = Column(Integer, index=True, server_default=text("0"))
    headstyle = Column(Integer, server_default=text("0"))
    linestyle = Column(Integer, server_default=text("0"))
    routestyle = Column(Integer, server_default=text("0"))
    isbold = Column(Integer, server_default=text("0"))
    linecolor = Column(Integer, server_default=text("0"))
    stereotype = Column(String(50))
    virtualinheritance = Column(CHAR(1))
    linkaccess = Column(String(50))
    pdata1 = Column(String(255), index=True)
    pdata2 = Column(Text)
    pdata3 = Column(String(255), index=True)
    pdata4 = Column(String(255))
    pdata5 = Column(Text, index=True)
    diagramid = Column(Integer, index=True, server_default=text("0"))
    ea_guid = Column(String(40), unique=True)
    sourceconstraint = Column(String(255))
    destconstraint = Column(String(255))
    sourceisnavigable = Column(Integer, server_default=text("0"))
    destisnavigable = Column(Integer, server_default=text("0"))
    isroot = Column(Integer, server_default=text("0"))
    isleaf = Column(Integer, server_default=text("0"))
    isspec = Column(Integer, server_default=text("0"))
    sourcechangeable = Column(String(12))
    destchangeable = Column(String(12))
    sourcets = Column(String(12))
    destts = Column(String(12))
    stateflags = Column(Text)
    actionflags = Column(String(255))
    issignal = Column(Integer, server_default=text("0"))
    isstimulus = Column(Integer, server_default=text("0"))
    dispatchaction = Column(String(255))
    target2 = Column(Integer)
    styleex = Column(Text, index=True)
    sourcestereotype = Column(String(255))
    deststereotype = Column(String(255))
    sourcestyle = Column(Text)
    deststyle = Column(Text)
    eventflags = Column(String(255))


class TAttributeconstraint(Base):
    __tablename__ = 't_attributeconstraints'

    object_id = Column(Integer, index=True, server_default=text("0"))
    Constraint = Column(String(255), primary_key=True, nullable=False)
    attname = Column(String(255))
    type = Column(String(255))
    notes = Column(Text)
    id = Column(Integer, primary_key=True, nullable=False)



# Default is JSON Schema Specification
generation_fields = {
    "default": {
        "bool": "boolean",
        "Date": "date",
        "dateTime": "date-time",
        "decimal": "number",
        "Decimal": "number",
        "enum": "enum",
        "int": "integer",
        "bigint": "integer",
        "Integer": "integer",
        "long": "integer",
        "String": "string",
        "str": "string",
    },
    "avro": {
        "bool": "boolean",
        "Date": "date",
        "dateTime": "date-time",
        "decimal": "number",
        "Decimal": "number",
        "enum": "enum",
        "int": "int",
        "integer": "int",
        "bigint": "int",
        "Integer": "int",
        "long": "int",
        "String": "string",
        "str": "string",
    },
    "spring data rest": {
        "boolean": "boolean",
        "date": "Date",
        "dateTime": "DateTime",
        "decimal": "Double",
        "enum": "String",
        "int": "int",
        "bigint": "int",
        "integer": "int",
        "long": "int",
        "string": "String",
    },
    "django": {
        "boolean": "BooleanField",
        "int": "IntegerField",
        "bigint": "BigIntegerField",
        "decimal": "DecimalField",
        "string": "CharField",
        "str": "CharField",
        "String": "CharField",
        "text": "TextField",
        "duration": "DurationField",
        "file": "FileField",
        "float": "FloatField",
        "date": "DateField",
        "dateTime": "DateTimeField",
        "datetime": "DateTimeField",
        "date_time": "DateTimeField",
        "json": "JSONField",
    },
    "marshmallow": {
        "boolean": "Boolean",
        "int": "Integer",
        "integer": "Integer",
        "bigint": "Integer",
        "decimal": "Decimal",
        "string": "String",
        "text": "Text",
        "duration": "Duration",
        "file": "File",
        "float": "Float",
        "date": "Date",
        "dateTime": "DateTime",
        "date_time": "DateTime",
    },
    "sqlalchemy": {
        "boolean": "Boolean",
        "int": "Integer",
        "integer": "Integer",
        "string": "String",
        "decimal": "Numeric",
        "Decimal": "Numeric",
        "float": "Float",
        "date": "Date",
        "dateTime": "DateTime",
        "date_time": "DateTime",
    },
    "python": {
        "Integer": "int",
        "String": "str",
        "string": "str",
        "Float": "float",
        "Numeric": "Decimal",
        "Boolean": "bool",
        "DECIMAL": "Decimal",
        "CHAR": "str",
        "TIMESTAMP": "datetime.timestamp",
        "DATE": "date",
        "INTEGER": "int",
        "VARCHAR": "str",
    },
    "ddl": {
        "Integer": "NUMBER",
        "int": "NUMBER",
        "String": "VARCHAR2",
        "string": "VARCHAR2",
        "str": "VARCHAR2",
        "decimal": "NUMBER",
    },
}



import re
from typing import List, Tuple
from lxml import etree
import logging
from decimal import Decimal

from mdg.config import settings
from mdg.uml import (
    UMLAssociation,
    UMLAttribute,
    UMLClass,
    UMLEnumeration,
    UMLInstance,
    UMLPackage,
    UMLAssociationType,
)


logger = logging.getLogger(__name__)

ns = {
    'uml': 'http://schema.omg.org/spec/UML/2.3',
    'xmi': 'http://schema.omg.org/spec/XMI/2.1',
}


def parse_uml() -> Tuple[UMLPackage, List[UMLInstance]]:
    """ Root package parser entry point.
    """
    test_cases: List[UMLInstance] = []

    # Parse into etree and grab root package
    logger.debug(f"loading etree from {settings['source']}")
    tree = etree.parse(settings['source'])
    model = tree.find('uml:Model', ns)
    root_package = model.xpath("//packagedElement[@name='%s']" % settings['root_package'], namespaces=ns)
    if len(root_package) == 0:
        raise ValueError("Root packaged element not found. Settings has:{}".format(settings['root_package']))
    element = root_package[0]

    # Find the element that is the root for models
    logger.info("Parsing models")
    model_element = element.xpath("//packagedElement[@name='%s']" % settings['model_package'], namespaces=ns)
    if len(model_element) == 0:
        raise ValueError("Model packaged element not found. Settings has:{}".format(settings['model_package']))
    model_element = model_element[0]

    # Create our root model UMLPackage and parse in 3 passes
    e_type = model_element.get('{%s}type' % ns['xmi'])
    if e_type == 'uml:Package':
        model_package = package_parse(model_element, tree, None)
        logger.debug("Parsing inheritance")
        # model_package_parse_inheritance(model_package)
        logger.debug("Parsing associations")
        # package_parse_associations(model_package, model_element, model_element)
        logger.debug("Sorting objects")
        # package_sort_classes(model_package)
    else:
        raise ValueError('Error - Non uml:Package element provided to packagedElement parser')

    if 'test_package' in settings.keys():
        logger.info("Parsing test cases")

        # Find the element that is the root for test data
        test_element = element.xpath("//packagedElement[@name='%s']" % settings['test_package'], namespaces=ns)
        if len(test_element) == 0:
            raise ValueError("Test packaged element not found. Settings has:{}".format(settings['test_package']))
        test_element = test_element[0]

        # Create our root test data UMLPackage and parse in 2 passes. Does not support inheritance
        e_type = test_element.get('{%s}type' % ns['xmi'])
        if e_type == 'uml:Package':
            test_package = package_parse(test_element, tree, None)
            package_parse_associations(test_package, test_element, test_element)
            logger.debug("Parsing inheritance")
            test_package_parse_inheritance(test_package, model_package)

            # With our test package parsed, we must return a list of instances instead of hierarchy of packages
            test_cases = parse_test_cases(test_package)
    return model_package, test_cases


def parse_test_cases(package) -> List[UMLInstance]:
    """ Looks through package hierarchy for instances with request or response stereotype
    and returns list of instances.
    :rtype: list<UMLInstance>
    """
    test_cases = []

    for instance in package.instances:
        if instance.stereotype in ['request', 'response']:
            test_cases.append(instance)

    for child in package.children:
        res = parse_test_cases(child)
        if res:
            test_cases += res

    return test_cases


def package_parse(element, root_element, parent_package):
    """ Extract package details, call class parser for classes and self parser for sub-packages.
    Associations are not done here, but in a 2nd pass using the parse_associations function.
    :param element:
    :param root_element:
    :return:
    :rtype: UMLPackage
    """
    name = element.get('name')
    id = element.get('{%s}id' % ns['xmi'])

    package = UMLPackage(id, name, parent_package)
    package._element = element
    package._root_element = root_element
    return package

    # Detail is Sparx specific
    # TODO: Put modelling tool in settings and use tool specific parser here
    detail = root_element.xpath("//element[@xmi:idref='%s']" % package.id, namespaces=ns)[0]
    properties = detail.find('properties')
    package.stereotype = properties.get('stereotype')
    if package.stereotype is not None:
        package.inherited_stereotypes.append([package.stereotype, package])

    package.documentation = properties.get('documentation')
    if package.documentation is None:
        package.documentation = ""

    project = detail.find('project')
    package.status = project.get('status')

    diagram_elements = root_element.xpath("//diagrams/diagram/model[@package='%s']" % package.id)
    for diagram_model in diagram_elements:
        diagram = diagram_model.getparent()
        package.diagrams.append(diagram.get('{%s}id' % ns['xmi']))

    logger.debug("Added UMLPackage {}".format(package.path))

    package_parse_children(element, package)
    return package


def package_parse_children(element, package):

    # Loop through all child elements and create nodes for classes and sub packages
    for child in element:
        e_type = child.get('{%s}type' % ns['xmi'])

        if e_type == 'uml:Package':
            pkg = package_parse(child, package._root_element, package)
            package.children.append(pkg)

        elif e_type == 'uml:Class':
            cls = class_parse(package, child, package._root_element)
            if cls.name is not None:
                package.classes.append(cls)

        elif e_type == 'uml:InstanceSpecification':
            ins = instance_parse(package, child, package._root_element)
            if ins.name is not None:
                package.instances.append(ins)

        elif e_type == 'uml:Enumeration':
            enumeration = enumeration_parse(package, child)
            if enumeration.name is not None:
                package.enumerations.append(enumeration)
            # print(f"Enum: {package.path}{enumeration}")


def package_parse_associations(package, element, root_element):
    """ Packages and classes should already have been parsed so now we link classes for each association.
    This gets messy as XMI output varies based on association type.
    This supports both un-specified and source to destination directional associations
    :param root_element:
    :param element: XML Element
    :type package: UMLPackage
    """
    for child in element:
        e_type = child.get('{%s}type' % ns['xmi'])
        e_id = child.get('{%s}id' % ns['xmi'])

        if e_type == 'uml:Association':
            assoc_source_id = None
            assoc_dest_id = None
            assoc_source_elem = None
            assoc_dest_elem = None

            for assoc in child:
                # If unspecified direction then both source and destination info are child elements within the
                # association
                assoc_type = assoc.get('{%s}type' % ns['xmi'])
                assoc_id = assoc.get('{%s}id' % ns['xmi'])
                if assoc_id is not None and assoc_type == 'uml:Property' and assoc_id[:8] == 'EAID_src':
                    assoc_source_elem = assoc
                    assoc_source_type_elem = assoc.find('type')
                    assoc_source_id = assoc_source_type_elem.get('{%s}idref' % ns['xmi'])
                if assoc_id is not None and assoc_type == 'uml:Property' and assoc_id[:8] == 'EAID_dst':
                    assoc_dest_elem = assoc
                    assoc_dest_type_elem = assoc.find('type')
                    assoc_dest_id = assoc_dest_type_elem.get('{%s}idref' % ns['xmi'])

            # If association direction is source to destination then
            # destination class info is found as an ownedAttribute in the source element
            if assoc_dest_id is None:
                for assoc in child:
                    if assoc.tag == 'memberEnd':
                        assoc_idref = assoc.get('{%s}idref' % ns['xmi'])
                        if assoc_idref[:8] == 'EAID_dst':
                            try:
                                assoc_dest_elem = \
                                    root_element.xpath("//ownedAttribute[@xmi:id='%s']" % assoc_idref,
                                                       namespaces=ns)[0]
                            except IndexError as e:
                                logger.warn("Failed to find member end association destination. Id: {}".format(assoc_idref))
                                raise e
                            assoc_dest_type_elem = assoc_dest_elem.find('type')
                            assoc_dest_id = assoc_dest_type_elem.get('{%s}idref' % ns['xmi'])

            # print("association: src id={} dest id={}".format(assoc_source_id,assoc_dest_id))
            # TODO: Raise error if we don't have a source and dest
            source = package.root_package.find_by_id(assoc_source_id)
            dest = package.root_package.find_by_id(assoc_dest_id)
            if source is not None \
                    and dest is not None \
                    and assoc_source_elem is not None \
                    and assoc_dest_elem is not None:
                association = association_parse(package, assoc_source_elem, assoc_dest_elem, source, dest)
                package.associations.append(association)
                logger.debug(f"Created {association.cardinality.name} {association.association_type.name} from {association.source_name} to {association.destination_name}")
            else:
                logger.warn("Unable to create association id={}".format(e_id))

    for package_child in package.children:
        element = element.xpath("//packagedElement[@xmi:id='%s']" % package_child.id, namespaces=ns)[0]
        package_parse_associations(package_child, element, root_element)


def model_package_parse_inheritance(package):
    """ Looks for classes which are specializations of a supertype and finds the correct object """
    for cls in package.classes:
        if cls.generalization_id is not None:
            cls.generalization = package.root_package.find_by_id(cls.generalization_id)
            if cls.generalization is None:
                logger.warn("Cannot find specialized class id={}".format(cls.generalization_id))
            else:
                cls.generalization.specialized_by.append(cls)
                if cls.id_attribute is None:
                    cls.id_attribute = cls.generalization.id_attribute

        for attr in cls.attributes:
            if attr.classification_id is not None:
                attr.classification = package.root_package.find_by_id(attr.classification_id)
                if attr.classification is None:
                    logger.warn("Cannot find expected classification for {} of attribute {}. Id={}".format(attr.dest_type, attr.name, attr.classification_id))

    for child in package.children:
        model_package_parse_inheritance(child)


def test_package_parse_inheritance(test_package, model_package):
    """ Links instances with the class they are instances of """

    for ins in test_package.instances:
        if ins.classification_id is not None:
            ins.classification = model_package.find_by_id(ins.classification_id)
            if ins.classification is None:
                logger.warn("Cannot find class which instance is from id={}".format(ins.classification_id))
            else:
                for attr in ins.attributes:
                    for cls_attr in ins.classification.attributes:
                        if attr.name == cls_attr.name:
                            attr.type = cls_attr.type
                            if attr.type.lower() in ['int', 'integer']:
                                attr.value = int(attr.value)
                            elif attr.type.lower() == ['float']:
                                attr.value = float(attr.value)
                            elif attr.type.lower() == ['decimal']:
                                attr.value = Decimal(attr.value)
                            break
        else:
            logger.warn("Instance object which is not from class id={}".format(ins.id))

    for child in test_package.children:
        test_package_parse_inheritance(child, model_package)


def package_sort_classes(package):
    # print(f"SORT {package.name}")
    unordered_list = package.classes
    unordered_composed_list = []
    for cls in unordered_list:
        if cls.composed_of != []:
            unordered_composed_list.append(cls)
            package.classes.remove(cls)
            # print(f"  REMOVE {cls.name}")

    # This is crap, do some propper sorting
    unordered_list = unordered_composed_list
    composed_list = []
    for cls1 in unordered_list:
        for cls2 in cls1.composed_of:
            if cls2 in unordered_composed_list:
                unordered_composed_list.remove(cls2)
                composed_list.append(cls2)

    composed_list += unordered_composed_list
    package.classes += composed_list

    for child in package.children:
        package_sort_classes(child)


def instance_parse(package, source_element, root):
    ins = UMLInstance(package, source_element.get('name'), source_element.get('{%s}id' % ns['xmi']))

    # Detail is sparx specific
    # TODO: Put modelling tool in settings and use tool specific parser here
    detail = root.xpath("//element[@xmi:idref='%s']" % ins.id, namespaces=ns)[0]
    properties = detail.find('properties')
    ins.stereotype = properties.get('stereotype')
    ins.documentation = properties.get('documentation')
    if ins.documentation is None:
        ins.documentation = ""

    project = detail.find('project')
    ins.status = project.get('status')

    # We need to link this instance to the class it is an instance of
    ins.classification_id = source_element.get('classifier')

    # Create attributes for each item found in the runstate
    # TODO: Change this to using an re
    extended_properties = detail.find('extendedProperties')
    if extended_properties is not None and extended_properties.get('runstate') is not None:
        run_state = extended_properties.get('runstate')
        vars = run_state.split('@ENDVAR;')
        for var in vars:
            if var != '':
                variable, value = (var.split(';')[1:3])
                attr = UMLAttribute(ins, variable.split('=')[1], value.split('=')[1])
                attr.value = value.split('=')[1]
                ins.attributes.append(attr)
    else:
        logger.info(f"No runstate found for UMLInstance {ins.name} | {ins.id}")
    logger.debug(f"Added UMLInstance {ins.name}")
    return ins


def association_parse(package, source_element, dest_element, source, dest):
    id = source_element.get('{%s}id' % ns['xmi'])
    association = UMLAssociation(package, source, dest, id)

    assoc_type = source_element.get("aggregation")
    if assoc_type == "composite":
        association.association_type = UMLAssociationType.COMPOSITION
        dest.composed_of.append(source)
    elif assoc_type == "shared":
        association.association_type = UMLAssociationType.AGGREGATION

    # Extract multiplicity for source
    source_lower = source_element.find('lowerValue')
    if source_lower is not None:
        source_lower = source_lower.get('value')
        if source_lower == '-1':
            source_lower = '*'
        source_upper = source_element.find('upperValue').get('value')
        if source_upper == '-1':
            source_upper = '*'
        association.source_multiplicity = (source_lower, source_upper)

    # Extract multiplicity for dest
    dest_lower = dest_element.find('lowerValue')
    if dest_lower is not None:
        dest_lower = dest_lower.get('value')
        if dest_lower == '-1':
            dest_lower = '*'
        dest_upper = dest_element.find('upperValue').get('value')
        if dest_upper == '-1':
            dest_upper = '*'
        association.destination_multiplicity = (dest_lower, dest_upper)

    # print('{} {} to {}'.format(association.source.name, association.association_type, association.destination.name))

    # If it's an association to or from a multiple then pluralize the name
    # TODO: Allow pluralized name to be specified in UML
    # Use opposing ends class name as attribute name for association
    association.destination_name = association.destination.name.lower()
    if association.destination_multiplicity[1] == '*':
        association.destination_name += 's'
    association.source_name = association.source.name.lower()
    if association.source_multiplicity[1] == '*':
        association.source_name += 's'

    # Allow explicit naming
    if dest_element.get('name') is not None:
        association.source_name = dest_element.get('name')
    if source_element.get('name') is not None:
        association.destination_name = source_element.get('name')

    # print('Assoc in {}: {} to {}: type = {}'.format(self.source.name, self.source_name, self.destination_name,
    # self.association_type) )
    return association


def enumeration_parse(package, element):
    enumeration = UMLEnumeration(package, element.get('name'), element.get('{%s}id' % ns['xmi']))

    # Loop through class elements children for values.
    for child in element:
        e_type = child.get('{%s}type' % ns['xmi'])
        if e_type == 'uml:EnumerationLiteral':
            enumeration.values.append(child.get('name'))
    logger.debug(f"Added UMLEnumeration {enumeration.name}")
    return enumeration


def class_parse(package, element, root):
    cls: UMLClass = UMLClass(package, element.get('name'), element.get('{%s}id' % ns['xmi']))
    if element.get('isAbstract') == 'true':
        cls.is_abstract = True
    else:
        cls.is_abstract = False

    # If the class is inherited from a superclass then get the ID. The actual object will be found in a separate pass
    # as it may not have been parsed yet
    generalization_element = element.find('generalization')
    if generalization_element is not None:
        cls.generalization_id = generalization_element.get('general')

    # Loop through class elements children for attributes.
    for child in element:
        e_type = child.get('{%s}type' % ns['xmi'])

        if e_type == 'uml:Property':
            # Associations will be done in a separate pass
            if child.get('association') is None and child.get('name') is not None:
                attr = attr_parse(cls, child, root)
                cls.attributes.append(attr)

    # Detail is sparx sprecific
    # TODO: Put modelling tool in settings and use tool specific parser here
    detail = root.xpath("//element[@xmi:idref='%s']" % cls.id, namespaces=ns)[0]
    properties = detail.find('properties')
    cls.alias = properties.get('alias')
    cls.documentation = properties.get('documentation')
    if cls.documentation is None:
        cls.documentation = ""

    project = detail.find('project')
    cls.status = project.get('status')

    # Get stereotypes, when multiple are provided only the first is found in the stereotype tag but all are found in
    # xrefs
    xrefs = detail.find('xrefs')
    value = xrefs.get('value')
    if value is not None:
        cls.stereotypes = re.findall('@STEREO;Name=(.*?);', value)

    logger.debug(f"Added UMLClass {cls.name}")
    return cls


def attr_parse(parent: UMLClass, element, root):
    attr = UMLAttribute(parent, element.get('name'), element.get('{%s}id' % ns['xmi']))

    attr.visibility = element.get('visibility')
    type_elem = element.find('type')
    if type_elem is not None:
        type_id = type_elem.get('{%s}idref' % ns['xmi'])
        if type_id[:4] == 'EAID':
            attr.classification_id = type_id
    else:
        logging.error(f"Attribute {attr.name} of class {parent} does not have a type")

    # Detail is sparx sprecific
    # TODO: Put modelling tool in settings and use tool specific parser here
    detail = root.xpath("//attribute[@xmi:idref='%s']" % attr.id, namespaces=ns)[0]
    properties = detail.find('properties')
    attr.set_type(properties.get('type'))
    doc = detail.find('documentation')
    doc = doc.get('value')
    if doc is not None:
        attr.documentation = doc

    alias_node = detail.find('style')
    attr.alias = alias_node.get('value')

    xrefs = detail.find('xrefs')
    if xrefs.get('value') is not None and 'NAME=isID' in xrefs.get('value'):
        attr.is_id = True
        attr.parent.id_attribute = attr
    else:
        attr.is_id = False

    stereotype = detail.find('stereotype')
    if stereotype is not None:
        attr.stereotype = stereotype.get('stereotype')

    constraints = detail.find('Constraints')
    if constraints is not None:
        for constraint in constraints:
            name = constraint.get('name')
            if name == 'unique':
                attr.is_unique = True
            elif name.startswith('length'):
                attr.length = int(name.split("=")[1])

    return attr



#!/usr/bin/python
import logging

from mdg.config import settings
from .sparx_xmi import parse_uml as sparx_parse_uml
from .drawio_xml import parse_uml as drawio_parse_uml
from .bouml_xmi import parse_uml as bouml_parse_uml
from .sparx_db import parse_uml as sparx_db_parse_uml
from .erwin_xmi import parse_uml as erwin_parse_uml


logger = logging.getLogger(__name__)


class ParseError(Exception):
    pass


def parse():
    """ Calls parser to turn model and tests into python native (see UML metamodel)
    """
    logger.debug("parse begin")

    # TODO: change to using a registration decorator
    PARSERS = {
        'sparx': sparx_parse_uml,
        'sparxdb': sparx_db_parse_uml,
        'drawio': drawio_parse_uml,
        'bouml': bouml_parse_uml,
        'erwin': erwin_parse_uml,
    }

    # Find the parser
    parser = None
    try:
        parser = PARSERS[settings['parser']]
        logger.debug(f"{settings['parser']} parser selected")
    except KeyError:
        raise ParseError("Error: Could not find parser. Settings must specify sparx, bouml or drawio")

    # Call the parser
    model_package, test_cases = parser()
    logger.info("Base Model Package: " + model_package.name)

    logger.debug("parse end")
    return model_package, test_cases



from typing import get_type_hints, Any, Tuple
from enum import Enum

DEFAULT_TYPES = [str, list, tuple, dict, bool, int, type(None)]


def obj_to_dict(obj: Any, base_dict={}) -> dict:
    """
    Creates nested dictionaries from input object
    """
    # Attributes to convert are specified in object definition type hints
    type_dict: dict = get_type_hints(type(obj))
    output = {}

    # Loop through all type hints on object and convert
    for attr_name in type_dict.keys():
        # Attribute must be set on instance
        if hasattr(obj, attr_name):
            value = getattr(obj, attr_name)

            # If the value is not a base type or enum then it's a class TODO: Actual check for class
            if type(value) not in DEFAULT_TYPES and not isinstance(value, Enum):
                # If the class is owned by this parent class then convert to a dict
                if attr_name in obj.Meta.owned_subobjects:
                    value = obj_to_dict(value)
                # If the class is not owned by this parent class then convert to a reference
                else:
                    value = getattr(value, value.Meta.id_field)

            # If value is a list then examine the type of the list and convert each element
            elif type(value) in [list, tuple]:
                new_value: Any = []
                for element in value:
                    # Check for base types as in non list mode above: TODO: Actual check for class
                    if type(element) not in DEFAULT_TYPES and not isinstance(element, Enum):
                        # Check for class ownership as above
                        if attr_name in obj.Meta.owned_subobjects:
                            new_value.append(obj_to_dict(element))
                        else:
                            new_value.append(getattr(element, element.Meta.id_field))
                    # Just a base type so use the value
                    else:
                        new_value.append(element)
                value = new_value

            elif isinstance(value, Enum):
                value = value.name

            # Value has been created so set the dictionary item
            output[attr_name] = value
    return output


def dict_to_obj(input: dict, base_object_class, references: dict = {}) -> object:
    obj, refs = dict_to_obj_pass1(input, base_object_class)
    dict_to_obj_pass2(obj, refs)
    return obj


def dict_to_obj_pass2(obj: Any, references: dict) -> None:
    """
    Pass 2 walks through objects and replaces attributes which are references with the actual class.
    """
    type_dict: dict = get_type_hints(type(obj))

    # Loop through all type hints on object
    for attr_name in type_dict.keys():
        # Walk through all attributes of object
        if hasattr(obj, attr_name):
            value = getattr(obj, attr_name)
            obj_attr_type = type_dict[attr_name]

            if type(value) == list:
                # We have a list if its supposed to be a list of classes then examine
                if obj_attr_type.__args__[0] not in DEFAULT_TYPES:
                    new_list = []
                    for item in value:
                        # If item is supposed to be a class but is actually a default type then must be reference
                        if type(item) in DEFAULT_TYPES:
                            new_list.append(references[value])
                        else:
                            dict_to_obj_pass2(item, references)
                            new_list.append(item)
                    setattr(obj, attr_name, new_list)

            elif obj_attr_type not in DEFAULT_TYPES:
                # If value is supposed to be a class but is actually a default type then must be reference
                if type(value) in DEFAULT_TYPES:
                    setattr(obj, attr_name, references[value])
                # If value is a class then parse it
                else:
                    dict_to_obj_pass2(value, references)


def dict_to_obj_pass1(input: dict, base_object_class, references: dict = {}) -> Tuple[Any, dict]:
    """
    Creates objects from input dictionary. Pass 1 keeps references and compiles a list of objects with ids used in pass 2.
    """
    # Instantiate the requested class & get the type hints
    obj = base_object_class()
    type_dict = get_type_hints(type(obj))

    for attr_name in type_dict.keys():
        if attr_name in input.keys():
            # The input has this item defined in the class so extract
            # dict_attr_type = type(input[attr_name])  # type_dict[attr_name].__origin__ TODO: use definition rather than input
            obj_attr_type = type_dict[attr_name]

            if type(input[attr_name]) != list:
                if obj_attr_type in DEFAULT_TYPES:
                    # Value definition is standard type so just set the value
                    setattr(obj, attr_name, input[attr_name])
                elif type(input[attr_name]) == dict:
                    # Not a standard type, represented as dict. Recurse to expand sub-object
                    new_obj, new_refs = dict_to_obj_pass1(input[attr_name], type_dict[attr_name], references)
                    setattr(obj, attr_name, new_obj)
                    # Store object in reference dictionary and add new object
                    references.update(new_refs)
                    references[getattr(new_obj, new_obj.Meta.id_field)] = new_obj
                else:
                    # If definition is class and dict value has string then must be a reference using id_field value from Meta
                    setattr(obj, attr_name, input[attr_name])
            else:
                # Input has a list of something, loop through each element and extract
                value = []
                item_type = type_dict[attr_name].__args__[0]
                for item in input[attr_name]:
                    if item_type in DEFAULT_TYPES:
                        # List of standard type so just add the value
                        value.append(item)
                    else:
                        # attr_class = type_dict[attr_name].__args__[0]
                        if type(item) == dict:
                            # Class definition has list of class instances and input has a dictionary so recurse to expand sub-object
                            new_obj, new_refs = dict_to_obj_pass1(item, item_type)
                            value.append(new_obj)
                            # Store object in reference dictionary and add new object
                            references.update(new_refs)
                            references[getattr(new_obj, new_obj.Meta.id_field)] = new_obj
                        else:
                            # If definition is class and dict value has string then must be a reference using id_field value from Meta
                            value.append(item)
                setattr(obj, attr_name, value)
    return obj, references



import unittest

from mdg.tools.case import camelcase, snakecase
from typing import List
from mdg.tools.io import obj_to_dict, dict_to_obj


class TestUMLModel(unittest.TestCase):
    def setUp(self):
        pass

    def test_camel(self):
        self.assertEqual("TestCase", camelcase("TestCase"))
        self.assertEqual("TestCase", camelcase("testCase"))
        self.assertEqual("TestCase", camelcase("test case"))
        self.assertEqual("TestCase", camelcase("Test Case"))
        self.assertEqual("TestCase", camelcase("Test_Case"))

    def test_snake(self):
        self.assertEqual("test_case", snakecase("testCase"))
        self.assertEqual("test_case", snakecase("test_case"))
        self.assertEqual("test_case", snakecase("TestCase"))
        self.assertEqual("test_case", snakecase("test case"))
        self.assertEqual("test_case", snakecase("Test Case"))
        self.assertEqual(None, snakecase(None))
        self.assertEqual("eftpos", snakecase("EFTPOS"))


class foo:
    id: int

    class Meta:
        id_field = 'id'


class blort:
    my_list: List[foo]
    basic_list: List[int]
    c: foo
    d: dict

    class Meta:
        owned_subobjects = 'my_list'


class TestIO(unittest.TestCase):
    def setUp(self):
        a = foo()
        a.id = 2

        self.b = blort()
        self.b.my_list = [a, ]
        self.b.c = a
        self.b.d = {"e": "f"}

    def test_to_dict(self):
        output = obj_to_dict(self.b)
        self.assertEqual(2, output['my_list'][0]['id'])

    def test_to_obj(self):
        input = {'c': {'id': 1}}
        obj = dict_to_obj(input, blort)
        self.assertEqual(1, obj.c.id)

        input = {'my_list': [{'id': 1}, ], 'c': 1, 'd': {"e": "f"}, 'basic_list': [1, 2, 3]}
        obj = dict_to_obj(input, blort)
        self.assertEqual(1, obj.c.id)
        self.assertEqual('f', obj.d['e'])
        self.assertEqual(2, obj.basic_list[1])
        self.assertEqual(1, obj.my_list[0].id)



import unittest

from mdg.uml import UMLClass, UMLPackage, UMLAssociation, Cardinality, UMLAssociationType, UMLAttribute, UMLEnumeration, SearchTypes


class TestUMLModel(unittest.TestCase):
    def setUp(self):
        self.root_package = UMLPackage("1", "root")
        child = UMLPackage("2", "child1", self.root_package)
        self.root_package.children.append(child)
        cls = UMLClass(child, "class1", "3")
        child.classes.append(cls)
        cls = UMLClass(child, "class2", "4")
        child.classes.append(cls)
        enum = UMLEnumeration(child, "enum1", "4")
        child.enumerations.append(enum)

    def test_find_package(self):
        res = self.root_package.find_by_id("2")
        self.assertEqual(UMLPackage, type(res))
        self.assertEqual("child1", res.name)

    def test_find_class(self):
        res = self.root_package.find_by_id("3", SearchTypes.CLASS)
        self.assertEqual(UMLClass, type(res))
        self.assertEqual("class1", res.name)

    def test_find_enumeration(self):
        # Check that we don't find the wrong thing
        res = self.root_package.find_by_id("3", SearchTypes.ENUM)
        self.assertEqual(None, res)
        # Check that we find the enum with the id=4, not the class with id=4
        res = self.root_package.find_by_id("4", SearchTypes.ENUM)
        self.assertEqual(UMLEnumeration, type(res))
        self.assertEqual("enum1", res.name)

    def test_string_to_multiplicity(self):
        assoc = UMLAssociation(self.root_package, self.root_package.children[0].classes[0], self.root_package.children[0].classes[0], 1)
        self.assertEqual(assoc.string_to_multiplicity("0..1"), ("0", "1"))
        self.assertEqual(assoc.string_to_multiplicity("0..*"), ("0", "*"))
        self.assertEqual(assoc.string_to_multiplicity("*..1"), ("*", "1"))

    def test_association_cardinality(self):
        assoc = UMLAssociation(self.root_package, self.root_package.children[0].classes[0], self.root_package.children[0].classes[0], 1)

        assoc.source_multiplicity = ("0", "1")
        assoc.destination_multiplicity = ("0", "*")
        self.assertEqual(assoc.cardinality, Cardinality.ONE_TO_MANY)

        assoc.source_multiplicity = ("0", "1")
        assoc.destination_multiplicity = ("0", "1")
        self.assertEqual(assoc.cardinality, Cardinality.ONE_TO_ONE)

        assoc.source_multiplicity = ("0", "*")
        assoc.destination_multiplicity = ("0", "1")
        self.assertEqual(assoc.cardinality, Cardinality.MANY_TO_ONE)

        assoc.source_multiplicity = ("0", "*")
        assoc.destination_multiplicity = ("0", "*")
        self.assertEqual(assoc.cardinality, Cardinality.MANY_TO_MANY)

    def test_association_type(self):
        assoc = UMLAssociation(self.root_package, self.root_package.children[0].classes[0], self.root_package.children[0].classes[0], 1)
        self.assertEqual(assoc.association_type, UMLAssociationType.ASSOCIATION)
        assoc = UMLAssociation(self.root_package, self.root_package.children[0].classes[0], self.root_package.children[0].classes[0], 1, UMLAssociationType.COMPOSITION)
        self.assertEqual(assoc.association_type, UMLAssociationType.COMPOSITION)

    def test_attribute_type(self):
        attr = UMLAttribute(None, "test", 123)
        attr.set_type('String')
        self.assertEqual(attr.dest_type, 'CharField')
        attr.set_type('String (123)')
        self.assertEqual(attr.dest_type, 'CharField')
        self.assertEqual(attr.length, 123)
        attr.set_type('decimal (12,2)')
        self.assertEqual(attr.dest_type, 'DecimalField')
        self.assertEqual(attr.precision, 12)
        self.assertEqual(attr.scale, 2)

    def test_attribute_get_type(self):
        attr = UMLAttribute(None, "test", 123)
        attr.set_type('String')
        self.assertEqual(attr.get_type('default'), 'string')

    def test_get_all_(self):
        self.assertEqual(len(self.root_package.get_all_classes()), 2)
        self.assertEqual(len(self.root_package.get_all_enums()), 1)



from typing import get_type_hints
from mdg.uml import UMLPackage, UMLClass, UMLAssociation, UMLEnumeration, UMLAttribute

schema: dict = {}

schema['UMLPackage'] = get_type_hints(UMLPackage)
schema['UMLClass'] = get_type_hints(UMLClass)
schema['UMLAssociation'] = get_type_hints(UMLAssociation)
schema['UMLEnumeration'] = get_type_hints(UMLEnumeration)
schema['UMLAttribute'] = get_type_hints(UMLAttribute)

print(schema)



from sqlalchemy import (
    CHAR,
    Column,
    # Date,
    DateTime,
    # Float,
    Index,
    Integer,
    # LargeBinary,
    # SmallInteger,
    String,
    # Table,
    Text,
    text,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.decl_api import declarative_base as _declarative_base

Base: _declarative_base = declarative_base()
metadata = Base.metadata


# class TAttribute(Base):
#     __tablename__ = 't_attribute'

#     object_id = Column(Integer, index=True, server_default=text("0"))
#     name = Column(String(255), index=True)
#     scope = Column(String(50))
#     stereotype = Column(String(50))
#     containment = Column(String(50))
#     isstatic = Column(Integer,server_default=text("0"))
#     iscollection = Column(Integer, server_default=text("0"))
#     isordered = Column(Integer, server_default=text("0"))
#     allowduplicates = Column(Integer, server_default=text("0"))
#     lowerbound = Column(String(50))
#     upperbound = Column(String(50))
#     container = Column(String(50))
#     notes = Column(Text)
#     derived = Column(CHAR(1))
#     id = Column(Integer, primary_key=True, server_default=text("nextval(('id_seq'::text)::regclass)"))
#     pos = Column(Integer)
#     genoption = Column(Text)
#     length = Column(Integer)
#     precision = Column(Integer)
#     scale = Column(Integer)
#     const = Column(Integer)
#     style = Column(String(255))
#     classifier = Column(String(50), index=True)
#     Default = Column(Text)
#     type = Column(String(255), index=True)
#     ea_guid = Column(String(50), unique=True)
#     styleex = Column(Text)


# class TAttributeconstraint(Base):
#     __tablename__ = 't_attributeconstraints'

#     object_id = Column(Integer, index=True, server_default=text("0"))
#     Constraint = Column(String(255), primary_key=True, nullable=False)
#     attname = Column(String(255))
#     type = Column(String(255))
#     notes = Column(Text)
#     id = Column(Integer, primary_key=True, nullable=False)


# class TAttributetag(Base):
#     __tablename__ = 't_attributetag'
#     __table_args__ = (
#         Index('ix_attributetag_elementidprop', 'elementid', 'property'),
#     )

#     propertyid = Column(Integer, primary_key=True, server_default=text("nextval(('propertyid_seq'::text)::regclass)"))
#     elementid = Column(Integer, index=True)
#     property = Column(String(255), index=True)
#     value = Column(String(255), index=True)
#     notes = Column(Text)
#     ea_guid = Column(String(40), index=True)


# class TAuthor(Base):
#     __tablename__ = 't_authors'

#     authorname = Column(String(255), primary_key=True)
#     roles = Column(String(255))
#     notes = Column(String(255))


# class TCardinality(Base):
#     __tablename__ = 't_cardinality'

#     cardinality = Column(String(50), primary_key=True)


# class TCategory(Base):
#     __tablename__ = 't_category'

#     categoryid = Column(Integer, primary_key=True, server_default=text("nextval(('categoryid_seq'::text)::regclass)"))
#     name = Column(String(255))
#     type = Column(String(255))
#     notes = Column(Text)


# class TClient(Base):
#     __tablename__ = 't_clients'

#     name = Column(String(255), primary_key=True)
#     organisation = Column(String(255))
#     phone1 = Column(String(50))
#     phone2 = Column(String(50))
#     mobile = Column(String(50))
#     fax = Column(String(50))
#     email = Column(String(50))
#     roles = Column(String(255))
#     notes = Column(String(255))


# class TComplexitytype(Base):
#     __tablename__ = 't_complexitytypes'

#     complexity = Column(String(50), primary_key=True)
#     numericweight = Column(Integer, index=True, server_default=text("0"))


# class TConnector(Base):
#     __tablename__ = 't_connector'
#     __table_args__ = (
#         Index('ix_connector_startobjidconnid', 'start_object_id', 'connector_id'),
#         Index('ix_connector_endobjidconnid', 'end_object_id', 'connector_id')
#     )

#     connector_id = Column(Integer, primary_key=True, server_default=text("nextval(('connector_id_seq'::text)::regclass)"))
#     name = Column(String(255))
#     direction = Column(String(50))
#     notes = Column(Text)
#     connector_type = Column(String(50), index=True)
#     subtype = Column(String(50), index=True)
#     sourcecard = Column(String(50))
#     sourceaccess = Column(String(50))
#     sourceelement = Column(String(50))
#     destcard = Column(String(50))
#     destaccess = Column(String(50))
#     destelement = Column(String(50))
#     sourcerole = Column(String(255))
#     sourceroletype = Column(String(50))
#     sourcerolenote = Column(Text)
#     sourcecontainment = Column(String(50))
#     sourceisaggregate = Column(Integer, server_default=text("0"))
#     sourceisordered = Column(Integer, server_default=text("0"))
#     sourcequalifier = Column(String(50))
#     destrole = Column(String(255))
#     destroletype = Column(String(50))
#     destrolenote = Column(Text)
#     destcontainment = Column(String(50))
#     destisaggregate = Column(Integer, server_default=text("0"))
#     destisordered = Column(Integer, server_default=text("0"))
#     destqualifier = Column(String(50))
#     start_object_id = Column(Integer, index=True, server_default=text("0"))
#     end_object_id = Column(Integer, index=True, server_default=text("0"))
#     top_start_label = Column(String(50))
#     top_mid_label = Column(String(50))
#     top_end_label = Column(String(50))
#     btm_start_label = Column(String(50))
#     btm_mid_label = Column(String(50))
#     btm_end_label = Column(String(50))
#     start_edge = Column(Integer, server_default=text("0"))
#     end_edge = Column(Integer, server_default=text("0"))
#     ptstartx = Column(Integer, server_default=text("0"))
#     ptstarty = Column(Integer, server_default=text("0"))
#     ptendx = Column(Integer, server_default=text("0"))
#     ptendy = Column(Integer, server_default=text("0"))
#     seqno = Column(Integer, index=True, server_default=text("0"))
#     headstyle = Column(Integer, server_default=text("0"))
#     linestyle = Column(Integer, server_default=text("0"))
#     routestyle = Column(Integer, server_default=text("0"))
#     isbold = Column(Integer, server_default=text("0"))
#     linecolor = Column(Integer, server_default=text("0"))
#     stereotype = Column(String(50))
#     virtualinheritance = Column(CHAR(1))
#     linkaccess = Column(String(50))
#     pdata1 = Column(String(255), index=True)
#     pdata2 = Column(Text)
#     pdata3 = Column(String(255), index=True)
#     pdata4 = Column(String(255))
#     pdata5 = Column(Text, index=True)
#     diagramid = Column(Integer, index=True, server_default=text("0"))
#     ea_guid = Column(String(40), unique=True)
#     sourceconstraint = Column(String(255))
#     destconstraint = Column(String(255))
#     sourceisnavigable = Column(Integer, server_default=text("0"))
#     destisnavigable = Column(Integer, server_default=text("0"))
#     isroot = Column(Integer, server_default=text("0"))
#     isleaf = Column(Integer, server_default=text("0"))
#     isspec = Column(Integer, server_default=text("0"))
#     sourcechangeable = Column(String(12))
#     destchangeable = Column(String(12))
#     sourcets = Column(String(12))
#     destts = Column(String(12))
#     stateflags = Column(Text)
#     actionflags = Column(String(255))
#     issignal = Column(Integer, server_default=text("0"))
#     isstimulus = Column(Integer, server_default=text("0"))
#     dispatchaction = Column(String(255))
#     target2 = Column(Integer)
#     styleex = Column(Text, index=True)
#     sourcestereotype = Column(String(255))
#     deststereotype = Column(String(255))
#     sourcestyle = Column(Text)
#     deststyle = Column(Text)
#     eventflags = Column(String(255))


# class TConnectorconstraint(Base):
#     __tablename__ = 't_connectorconstraint'

#     connectorid = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     Constraint = Column(String(255), primary_key=True, nullable=False, index=True)
#     constrainttype = Column(String(50))
#     notes = Column(Text)


# class TConnectortag(Base):
#     __tablename__ = 't_connectortag'

#     propertyid = Column(Integer, primary_key=True, server_default=text("nextval(('propertyid_seq'::text)::regclass)"))
#     elementid = Column(Integer, index=True)
#     property = Column(String(255), index=True)
#     value = Column(String(255), index=True)
#     notes = Column(Text)
#     ea_guid = Column(String(40), index=True)


# class TConnectortype(Base):
#     __tablename__ = 't_connectortypes'

#     connector_type = Column(String(50), primary_key=True)
#     description = Column(String(50))


# class TConstant(Base):
#     __tablename__ = 't_constants'

#     constantname = Column(String(50), primary_key=True)
#     constantvalue = Column(String(255))


# class TConstrainttype(Base):
#     __tablename__ = 't_constrainttypes'

#     Constraint = Column(String(16), primary_key=True)
#     description = Column(String(50))
#     notes = Column(Text)


# class TDatatype(Base):
#     __tablename__ = 't_datatypes'

#     type = Column(String(50))
#     productname = Column(String(50), index=True)
#     datatype = Column(String(50), index=True)
#     size = Column(Integer)
#     maxlen = Column(Integer)
#     maxprec = Column(Integer)
#     maxscale = Column(Integer, server_default=text("0"))
#     defaultlen = Column(Integer)
#     defaultprec = Column(Integer)
#     defaultscale = Column(Integer)
#     User = Column(Integer)
#     pdata1 = Column(String(255))
#     pdata2 = Column(String(255))
#     pdata3 = Column(String(255))
#     pdata4 = Column(String(255))
#     haslength = Column(String(50))
#     generictype = Column(String(255))
#     datatypeid = Column(Integer, primary_key=True, server_default=text("nextval(('datatypeid_seq'::text)::regclass)"))


# class TDiagram(Base):
#     __tablename__ = 't_diagram'

#     diagram_id = Column(Integer, primary_key=True, server_default=text("nextval(('diagram_id_seq'::text)::regclass)"))
#     package_id = Column(Integer, index=True, server_default=text("1"))
#     parentid = Column(Integer, index=True, server_default=text("0"))
#     diagram_type = Column(String(50), index=True)
#     name = Column(String(255))
#     version = Column(String(50), server_default=text("'1.0'::character varying"))
#     author = Column(String(255))
#     showdetails = Column(Integer, server_default=text("0"))
#     notes = Column(Text)
#     stereotype = Column(String(50))
#     attpub = Column(Integer, nullable=False, server_default=text("1"))
#     attpri = Column(Integer, nullable=False, server_default=text("1"))
#     attpro = Column(Integer, nullable=False, server_default=text("1"))
#     orientation = Column(CHAR(1), server_default=text("'P'::bpchar"))
#     cx = Column(Integer, server_default=text("0"))
#     cy = Column(Integer, server_default=text("0"))
#     scale = Column(Integer, server_default=text("100"))
#     createddate = Column(DateTime, server_default=text("now()"))
#     modifieddate = Column(DateTime, server_default=text("now()"))
#     htmlpath = Column(String(255))
#     showforeign = Column(Integer, nullable=False, server_default=text("1"))
#     showborder = Column(Integer, nullable=False, server_default=text("1"))
#     showpackagecontents = Column(Integer, nullable=False, server_default=text("1"))
#     pdata = Column(String(255))
#     locked = Column(Integer, nullable=False, server_default=text("0"))
#     ea_guid = Column(String(40), unique=True)
#     tpos = Column(Integer)
#     swimlanes = Column(String(255))
#     styleex = Column(Text)


# class TDiagramlink(Base):
#     __tablename__ = 't_diagramlinks'

#     diagramid = Column(Integer, index=True)
#     connectorid = Column(Integer, index=True)
#     geometry = Column(Text)
#     style = Column(String(255))
#     hidden = Column(Integer, nullable=False, server_default=text("0"))
#     path = Column(String(255))
#     instance_id = Column(Integer, primary_key=True, server_default=text("nextval(('instance_id_seq'::text)::regclass)"))


# class TDiagramobject(Base):
#     __tablename__ = 't_diagramobjects'

#     diagram_id = Column(Integer, index=True, server_default=text("0"))
#     object_id = Column(Integer, index=True, server_default=text("0"))
#     recttop = Column(Integer, server_default=text("0"))
#     rectleft = Column(Integer, server_default=text("0"))
#     rectright = Column(Integer, server_default=text("0"))
#     rectbottom = Column(Integer, server_default=text("0"))
#     sequence = Column(Integer, index=True, server_default=text("0"))
#     objectstyle = Column(String(255))
#     instance_id = Column(Integer, primary_key=True, server_default=text("nextval(('instance_id_seq'::text)::regclass)"))


# class TDiagramtype(Base):
#     __tablename__ = 't_diagramtypes'

#     diagram_type = Column(String(50), primary_key=True)
#     name = Column(String(255))
#     package_id = Column(Integer, index=True, server_default=text("0"))


# class TDocument(Base):
#     __tablename__ = 't_document'

#     docid = Column(String(40), primary_key=True)
#     docname = Column(String(100))
#     notes = Column(String(255))
#     style = Column(String(255), index=True)
#     elementid = Column(String(40), nullable=False, index=True)
#     elementtype = Column(String(50), nullable=False)
#     strcontent = Column(Text)
#     bincontent = Column(LargeBinary)
#     doctype = Column(String(100), index=True)
#     author = Column(String(255), index=True)
#     version = Column(String(50), index=True)
#     isactive = Column(Integer, server_default=text("1"))
#     sequence = Column(Integer, server_default=text("0"))
#     docdate = Column(DateTime)


# class TEcf(Base):
#     __tablename__ = 't_ecf'

#     ecfid = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     weight = Column(Float, index=True, server_default=text("1"))
#     value = Column(Float, server_default=text("0"))
#     notes = Column(String(255))


# class TEfforttype(Base):
#     __tablename__ = 't_efforttypes'

#     efforttype = Column(String(12), primary_key=True)
#     description = Column(String(255))
#     numericweight = Column(Float, index=True, server_default=text("0"))
#     notes = Column(String(255))


# class TFile(Base):
#     __tablename__ = 't_files'
#     __table_args__ = (
#         Index('ix_files_appliestoname', 'appliesto', 'name'),
#         Index('ix_files_appliestonamedatesize', 'appliesto', 'name', 'filedate', 'filesize'),
#         Index('ix_files_appliestonamesize', 'appliesto', 'name', 'filesize')
#     )

#     fileid = Column(String(50), primary_key=True)
#     appliesto = Column(String(50), nullable=False, index=True)
#     category = Column(String(100), nullable=False, index=True)
#     name = Column(String(150), nullable=False, index=True)
#     file = Column(String(255))
#     notes = Column(Text)
#     filedate = Column(DateTime)
#     filesize = Column(Integer, index=True)


# t_t_genopt = Table(
#     't_genopt', metadata,
#     Column('appliesto', String(12)),
#     Column('option', Text)
# )


# class TGlossary(Base):
#     __tablename__ = 't_glossary'

#     term = Column(String(255), index=True)
#     type = Column(String(255), index=True)
#     meaning = Column(Text)
#     glossaryid = Column(Integer, primary_key=True, server_default=text("nextval(('glossaryid_seq'::text)::regclass)"))


# t_t_html = Table(
#     't_html', metadata,
#     Column('type', String(50)),
#     Column('template', Text)
# )


# class TImage(Base):
#     __tablename__ = 't_image'

#     imageid = Column(Integer, primary_key=True, server_default=text("nextval(('imageid_seq'::text)::regclass)"))
#     name = Column(String(255))
#     type = Column(String(255))
#     image = Column(LargeBinary)


# t_t_implement = Table(
#     't_implement', metadata,
#     Column('type', String(50))
# )


# class TIssue(Base):
#     __tablename__ = 't_issues'

#     issue = Column(String(255))
#     issuedate = Column(DateTime)
#     owner = Column(String(255))
#     status = Column(String(50))
#     notes = Column(Text)
#     resolver = Column(String(255))
#     dateresolved = Column(DateTime)
#     resolution = Column(Text)
#     issueid = Column(Integer, primary_key=True, server_default=text("nextval(('issueid_seq'::text)::regclass)"))
#     category = Column(String(255))
#     priority = Column(String(50))
#     severity = Column(String(50))
#     issuetype = Column(String(100))


# class TList(Base):
#     __tablename__ = 't_lists'

#     listid = Column(String(50), primary_key=True)
#     category = Column(String(100), nullable=False)
#     name = Column(String(150), nullable=False)
#     nval = Column(Integer)
#     notes = Column(Text)


# class TMainttype(Base):
#     __tablename__ = 't_mainttypes'

#     mainttype = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


# class TMethod(Base):
#     __tablename__ = 't_method'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     name = Column(String(255), primary_key=True, nullable=False)
#     scope = Column(String(50))
#     type = Column(String(50))


# class TMetrictype(Base):
#     __tablename__ = 't_metrictypes'

#     metric = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


class TObject(Base):
    __tablename__ = 't_object'
    __table_args__ = (
        Index('ix_object_eaguidclassifier', 'ea_guid', 'classifier'),
        Index('ix_object_pkgidpdata1class', 'package_id', 'pdata1', 'classifier')
    )

    object_id = Column(Integer, primary_key=True, server_default=text("nextval(('object_id_seq'::text)::regclass)"))
    object_type = Column(String(255), index=True)
    diagram_id = Column(Integer, server_default=text("0"))
    name = Column(String(255), index=True)
    alias = Column(String(255))
    author = Column(String(255))
    version = Column(String(50), server_default=text("'1.0'::character varying"))
    note = Column(Text)
    package_id = Column(Integer, index=True, server_default=text("0"))
    stereotype = Column(String(255))
    ntype = Column(Integer, index=True, server_default=text("0"))
    complexity = Column(String(50), server_default=text("'2'::character varying"))
    effort = Column(Integer, server_default=text("0"))
    style = Column(String(255))
    backcolor = Column(Integer, server_default=text("0"))
    borderstyle = Column(Integer, server_default=text("0"))
    borderwidth = Column(Integer, server_default=text("0"))
    fontcolor = Column(Integer, server_default=text("0"))
    bordercolor = Column(Integer, server_default=text("0"))
    createddate = Column(DateTime, server_default=text("now()"))
    modifieddate = Column(DateTime, server_default=text("now()"))
    status = Column(String(50))
    abstract = Column(CHAR(1))
    tagged = Column(Integer, server_default=text("0"))
    pdata1 = Column(String(255), index=True)
    pdata2 = Column(Text, index=True)
    pdata3 = Column(Text, index=True)
    pdata4 = Column(Text, index=True)
    pdata5 = Column(String(255), index=True)
    concurrency = Column(String(50))
    visibility = Column(String(50))
    persistence = Column(String(50))
    cardinality = Column(String(50))
    gentype = Column(String(50))
    genfile = Column(String(255))
    header1 = Column(Text)
    header2 = Column(Text)
    phase = Column(String(50))
    scope = Column(String(25))
    genoption = Column(Text)
    genlinks = Column(Text, index=True)
    classifier = Column(Integer, index=True)
    ea_guid = Column(String(40), unique=True)
    parentid = Column(Integer, index=True)
    runstate = Column(Text)
    classifier_guid = Column(String(40), index=True)
    tpos = Column(Integer)
    isroot = Column(Integer, server_default=text("0"))
    isleaf = Column(Integer, server_default=text("0"))
    isspec = Column(Integer, server_default=text("0"))
    isactive = Column(Integer, server_default=text("0"))
    stateflags = Column(String(255))
    packageflags = Column(String(255))
    multiplicity = Column(String(50))
    styleex = Column(Text)
    actionflags = Column(String(255))
    eventflags = Column(String(255), index=True)


# class TObjectconstraint(Base):
#     __tablename__ = 't_objectconstraint'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     Constraint = Column(String(255), primary_key=True, nullable=False, index=True)
#     constrainttype = Column(String(30), primary_key=True, nullable=False)
#     weight = Column(Float, server_default=text("0"))
#     notes = Column(Text)
#     status = Column(String(50))


# class TObjecteffort(Base):
#     __tablename__ = 't_objecteffort'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     effort = Column(String(255), primary_key=True, nullable=False)
#     efforttype = Column(String(12))
#     evalue = Column(Float, server_default=text("0"))
#     notes = Column(Text)


# class TObjectfile(Base):
#     __tablename__ = 't_objectfiles'

#     object_id = Column(Integer, primary_key=True, nullable=False, server_default=text("0"))
#     filename = Column(String(255), primary_key=True, nullable=False)
#     type = Column(String(50))
#     note = Column(Text)
#     filesize = Column(String(255))
#     filedate = Column(String(255))


# class TObjectmetric(Base):
#     __tablename__ = 't_objectmetrics'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     metric = Column(String(255), primary_key=True, nullable=False, index=True)
#     metrictype = Column(String(12), index=True)
#     evalue = Column(Float, server_default=text("1"))
#     notes = Column(Text)


# class TObjectproblem(Base):
#     __tablename__ = 't_objectproblems'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     problem = Column(String(255), primary_key=True, nullable=False, index=True)
#     problemtype = Column(String(255), primary_key=True, nullable=False)
#     datereported = Column(Date)
#     status = Column(String(50))
#     problemnotes = Column(Text)
#     reportedby = Column(String(255))
#     resolvedby = Column(String(255))
#     dateresolved = Column(Date)
#     version = Column(String(50))
#     resolvernotes = Column(Text)
#     priority = Column(String(50))
#     severity = Column(String(50))


# class TObjectproperty(Base):
#     __tablename__ = 't_objectproperties'
#     __table_args__ = (
#         Index('ix_objectproperties_objidprop', 'object_id', 'property'),
#     )

#     propertyid = Column(Integer, primary_key=True, server_default=text("nextval(('propertyid_seq'::text)::regclass)"))
#     object_id = Column(Integer, index=True, server_default=text("0"))
#     property = Column(String(255), index=True)
#     value = Column(String(255), index=True)
#     notes = Column(Text)
#     ea_guid = Column(String(40), index=True)


# class TObjectrequire(Base):
#     __tablename__ = 't_objectrequires'

#     reqid = Column(Integer, primary_key=True, server_default=text("nextval(('reqid_seq'::text)::regclass)"))
#     object_id = Column(Integer, index=True, server_default=text("0"))
#     requirement = Column(String(255), index=True)
#     reqtype = Column(String(255))
#     status = Column(String(50))
#     notes = Column(Text)
#     stability = Column(String(50))
#     difficulty = Column(String(50))
#     priority = Column(String(50))
#     lastupdate = Column(DateTime, server_default=text("now()"))


# class TObjectresource(Base):
#     __tablename__ = 't_objectresource'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     resource = Column(String(255), primary_key=True, nullable=False)
#     role = Column(String(255), primary_key=True, nullable=False)
#     time = Column(Float, server_default=text("0"))
#     notes = Column(Text)
#     percentcomplete = Column(SmallInteger, server_default=text("0"))
#     datestart = Column(Date)
#     dateend = Column(Date)
#     history = Column(Text)
#     expectedhours = Column(Integer)
#     actualhours = Column(Integer)


# class TObjectrisk(Base):
#     __tablename__ = 't_objectrisks'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     risk = Column(String(255), primary_key=True, nullable=False)
#     risktype = Column(String(12))
#     evalue = Column(Float, server_default=text("0"))
#     notes = Column(Text)


# class TObjectscenario(Base):
#     __tablename__ = 't_objectscenarios'
#     __table_args__ = (
#         Index('ix_objectscenarios_oidevalscen', 'object_id', 'evalue', 'scenario'),
#     )

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     scenario = Column(String(255), primary_key=True, nullable=False)
#     scenariotype = Column(String(12))
#     evalue = Column(Float, server_default=text("0"))
#     notes = Column(Text)
#     ea_guid = Column(String(40), nullable=False)
#     xmlcontent = Column(Text)


# class TObjecttest(Base):
#     __tablename__ = 't_objecttests'
#     __table_args__ = (
#         Index('ix_objecttexts_objidtclassstat', 'object_id', 'testclass', 'status'),
#     )

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     test = Column(String(255), primary_key=True, nullable=False)
#     testclass = Column(Integer, primary_key=True, nullable=False, server_default=text("0"))
#     testtype = Column(String(50))
#     notes = Column(Text)
#     inputdata = Column(Text)
#     acceptancecriteria = Column(Text)
#     status = Column(String(32))
#     daterun = Column(Date, server_default=text("now()"))
#     results = Column(Text)
#     runby = Column(String(255))
#     checkby = Column(String(255))


# class TObjecttrx(Base):
#     __tablename__ = 't_objecttrx'

#     object_id = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     trx = Column(String(255), primary_key=True, nullable=False, index=True)
#     trxtype = Column(String(12), primary_key=True, nullable=False)
#     weight = Column(Float, server_default=text("0"))
#     notes = Column(Text)


# class TObjecttype(Base):
#     __tablename__ = 't_objecttypes'

#     object_type = Column(String(50), primary_key=True)
#     description = Column(String(255))
#     designobject = Column(Integer, nullable=False, server_default=text("0"))
#     imageid = Column(SmallInteger, index=True, server_default=text("0"))


# t_t_ocf = Table(
#     't_ocf', metadata,
#     Column('objecttype', String(50)),
#     Column('complexityweight', Float, server_default=text("0"))
# )


# class TOperation(Base):
#     __tablename__ = 't_operation'
#     __table_args__ = (
#         Index('ix_operation_objectidopid', 'object_id', 'operationid'),
#     )

#     operationid = Column(Integer, primary_key=True, server_default=text("nextval(('operationid_seq'::text)::regclass)"))
#     object_id = Column(Integer, index=True, server_default=text("0"))
#     name = Column(String(255), index=True)
#     scope = Column(String(50))
#     type = Column(String(255))
#     returnarray = Column(CHAR(1))
#     stereotype = Column(String(50))
#     isstatic = Column(CHAR(1))
#     concurrency = Column(String(50))
#     notes = Column(Text)
#     behaviour = Column(Text)
#     abstract = Column(CHAR(1))
#     genoption = Column(Text)
#     synchronized = Column(CHAR(1))
#     pos = Column(Integer)
#     const = Column(Integer)
#     style = Column(String(255))
#     pure = Column(Integer, nullable=False, server_default=text("0"))
#     throws = Column(String(255))
#     classifier = Column(String(50), index=True)
#     code = Column(Text)
#     isroot = Column(Integer, server_default=text("0"))
#     isleaf = Column(Integer, server_default=text("0"))
#     isquery = Column(Integer, server_default=text("0"))
#     stateflags = Column(String(255))
#     ea_guid = Column(String(50), unique=True)
#     styleex = Column(Text)


# class TOperationparam(Base):
#     __tablename__ = 't_operationparams'
#     __table_args__ = (
#         Index('ix_operationparams_opidpos', 'operationid', 'pos'),
#     )

#     operationid = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     name = Column(String(255), primary_key=True, nullable=False)
#     type = Column(String(255))
#     Default = Column(String(255))
#     notes = Column(Text)
#     pos = Column(Integer)
#     const = Column(Integer, server_default=text("0"))
#     style = Column(String(255))
#     kind = Column(String(12))
#     classifier = Column(String(50), index=True)
#     ea_guid = Column(String(50), unique=True)
#     styleex = Column(Text)


# class TOperationpost(Base):
#     __tablename__ = 't_operationposts'

#     operationid = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     postcondition = Column(String(255), primary_key=True, nullable=False)
#     type = Column(String(255))
#     notes = Column(Text)


# class TOperationpre(Base):
#     __tablename__ = 't_operationpres'

#     operationid = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     precondition = Column(String(255), primary_key=True, nullable=False)
#     type = Column(String(50))
#     notes = Column(Text)


# class TOperationtag(Base):
#     __tablename__ = 't_operationtag'
#     __table_args__ = (
#         Index('ix_operationtag_elementidprop', 'elementid', 'property'),
#     )

#     propertyid = Column(Integer, primary_key=True, server_default=text("nextval(('propertyid_seq'::text)::regclass)"))
#     elementid = Column(Integer, index=True)
#     property = Column(String(255), index=True)
#     value = Column(String(255), index=True)
#     notes = Column(Text)
#     ea_guid = Column(String(40), index=True)


class TPackage(Base):
    __tablename__ = 't_package'

    package_id = Column(Integer, primary_key=True, server_default=text("nextval(('package_id_seq'::text)::regclass)"))
    name = Column(String(255), index=True)
    parent_id = Column(Integer, index=True, server_default=text("0"))
    createddate = Column(DateTime, server_default=text("now()"))
    modifieddate = Column(DateTime, server_default=text("now()"))
    notes = Column(Text)
    ea_guid = Column(String(40), unique=True)
    xmlpath = Column(String(255))
    iscontrolled = Column(Integer, server_default=text("0"))
    lastloaddate = Column(DateTime)
    lastsavedate = Column(DateTime)
    version = Column(String(50))
    protected = Column(Integer, server_default=text("0"))
    pkgowner = Column(String(255))
    umlversion = Column(String(50))
    usedtd = Column(Integer, server_default=text("0"))
    logxml = Column(Integer, server_default=text("0"))
    codepath = Column(String(255))
    namespace = Column(String(50))
    tpos = Column(Integer)
    packageflags = Column(String(255))
    batchsave = Column(Integer)
    batchload = Column(Integer)


# class TPalette(Base):
#     __tablename__ = 't_palette'

#     paletteid = Column(Integer, primary_key=True, server_default=text("nextval(('paletteid_seq'::text)::regclass)"))
#     name = Column(String(255))
#     type = Column(String(255))


# t_t_paletteitem = Table(
#     't_paletteitem', metadata,
#     Column('paletteid', Integer),
#     Column('itemid', Integer)
# )


# class TPhase(Base):
#     __tablename__ = 't_phase'

#     phaseid = Column(String(40), primary_key=True)
#     phasename = Column(String(100), nullable=False)
#     phasenotes = Column(Text)
#     phasestyle = Column(String(255))
#     startdate = Column(DateTime)
#     enddate = Column(DateTime)
#     phasecontent = Column(Text)


# class TPrimitive(Base):
#     __tablename__ = 't_primitives'

#     datatype = Column(String(50), primary_key=True)
#     description = Column(String(50))


# class TProblemtype(Base):
#     __tablename__ = 't_problemtypes'

#     problemtype = Column(String(12), primary_key=True)
#     description = Column(String(255))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


# class TProjectrole(Base):
#     __tablename__ = 't_projectroles'

#     role = Column(String(255), primary_key=True)
#     description = Column(String(255))
#     notes = Column(Text)


# class TPropertytype(Base):
#     __tablename__ = 't_propertytypes'

#     property = Column(String(50), primary_key=True)
#     description = Column(String(50))
#     notes = Column(Text)


# class TRequiretype(Base):
#     __tablename__ = 't_requiretypes'

#     requirement = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


# class TResource(Base):
#     __tablename__ = 't_resources'

#     name = Column(String(255), primary_key=True)
#     organisation = Column(String(255))
#     phone1 = Column(String(50))
#     phone2 = Column(String(50))
#     mobile = Column(String(50))
#     fax = Column(String(50))
#     email = Column(String(255))
#     roles = Column(String(255))
#     notes = Column(String(255))


# class TRisktype(Base):
#     __tablename__ = 't_risktypes'

#     risk = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("0"))
#     notes = Column(String(255))


# class TRoleconstraint(Base):
#     __tablename__ = 't_roleconstraint'

#     connectorid = Column(Integer, primary_key=True, nullable=False, index=True, server_default=text("0"))
#     Constraint = Column(String(255), primary_key=True, nullable=False, index=True)
#     connectorend = Column(String(50), primary_key=True, nullable=False)
#     constrainttype = Column(String(12), primary_key=True, nullable=False)
#     notes = Column(Text)


# t_t_rtf = Table(
#     't_rtf', metadata,
#     Column('type', String(50)),
#     Column('template', Text)
# )


# class TRtfreport(Base):
#     __tablename__ = 't_rtfreport'

#     templateid = Column(String(200), primary_key=True)
#     rootpackage = Column(Integer, server_default=text("0"))
#     filename = Column(String(255))
#     details = Column(Integer, server_default=text("0"))
#     processchildren = Column(Integer, server_default=text("0"))
#     showdiagrams = Column(Integer, server_default=text("0"))
#     heading = Column(String(255))
#     requirements = Column(Integer, server_default=text("0"))
#     associations = Column(Integer, server_default=text("0"))
#     scenarios = Column(Integer, server_default=text("0"))
#     childdiagrams = Column(Integer, server_default=text("0"))
#     attributes = Column(Integer, server_default=text("0"))
#     methods = Column(Integer, server_default=text("0"))
#     imagetype = Column(Integer, server_default=text("0"))
#     paging = Column(Integer, server_default=text("0"))
#     intro = Column(Text)
#     resources = Column(Integer, server_default=text("1"))
#     constraints = Column(Integer, server_default=text("1"))
#     tagged = Column(Integer, server_default=text("0"))
#     showtag = Column(Integer, server_default=text("0"))
#     showalias = Column(Integer, server_default=text("0"))
#     pdata1 = Column(String(255))
#     pdata2 = Column(String(255))
#     pdata3 = Column(String(255))
#     pdata4 = Column(Text)


# class TRule(Base):
#     __tablename__ = 't_rules'

#     ruleid = Column(String(50), primary_key=True)
#     rulename = Column(String(255), nullable=False)
#     ruletype = Column(String(255), nullable=False)
#     ruleactive = Column(Integer)
#     errormsg = Column(String(255))
#     flags = Column(String(255))
#     ruleocl = Column(Text)
#     rulexml = Column(Text)
#     notes = Column(Text)


# class TScenariotype(Base):
#     __tablename__ = 't_scenariotypes'

#     scenariotype = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


# class TScript(Base):
#     __tablename__ = 't_script'

#     scriptid = Column(Integer, primary_key=True, server_default=text("nextval(('scriptid_seq'::text)::regclass)"))
#     scriptcategory = Column(String(100), index=True)
#     scriptname = Column(String(150))
#     scriptauthor = Column(String(255))
#     notes = Column(Text)
#     script = Column(Text)


# class TSecgroup(Base):
#     __tablename__ = 't_secgroup'

#     groupid = Column(String(40), primary_key=True)
#     groupname = Column(String(32), nullable=False)
#     description = Column(String(100))


# class TSecgrouppermission(Base):
#     __tablename__ = 't_secgrouppermission'

#     groupid = Column(String(40), primary_key=True, nullable=False)
#     permissionid = Column(Integer, primary_key=True, nullable=False)


# class TSeclock(Base):
#     __tablename__ = 't_seclocks'
#     __table_args__ = (
#         Index('ix_seclocks_entityidusrtype', 'entityid', 'userid', 'entitytype'),
#         Index('ix_seclocks_entityiduserid', 'entityid', 'userid'),
#         Index('ix_seclocks_entityidtype', 'entityid', 'entitytype'),
#         Index('ix_seclocks_groupiduserid', 'groupid', 'userid')
#     )

#     userid = Column(String(40), nullable=False)
#     groupid = Column(String(40), index=True)
#     entitytype = Column(String(32), nullable=False)
#     entityid = Column(String(40), primary_key=True)
#     timestamp = Column(DateTime, nullable=False)
#     locktype = Column(String(255))


# class TSecpermission(Base):
#     __tablename__ = 't_secpermission'

#     permissionid = Column(Integer, primary_key=True)
#     permissionname = Column(String(50), nullable=False)


# class TSecpolicy(Base):
#     __tablename__ = 't_secpolicies'

#     property = Column(String(100), primary_key=True)
#     value = Column(String(255), nullable=False)


# class TSecuser(Base):
#     __tablename__ = 't_secuser'

#     userid = Column(String(40), primary_key=True)
#     userlogin = Column(String(255), nullable=False)
#     firstname = Column(String(50), nullable=False)
#     surname = Column(String(50), nullable=False)
#     department = Column(String(50))
#     password = Column(String(12))


# class TSecusergroup(Base):
#     __tablename__ = 't_secusergroup'

#     userid = Column(String(40), primary_key=True, nullable=False)
#     groupid = Column(String(40), primary_key=True, nullable=False)


# class TSecuserpermission(Base):
#     __tablename__ = 't_secuserpermission'

#     userid = Column(String(40), primary_key=True, nullable=False)
#     permissionid = Column(Integer, primary_key=True, nullable=False)


# class TSnapshot(Base):
#     __tablename__ = 't_snapshot'

#     snapshotid = Column(String(40), primary_key=True)
#     seriesid = Column(String(40), nullable=False, index=True)
#     position = Column(Integer, index=True)
#     snapshotname = Column(String(100))
#     notes = Column(Text)
#     style = Column(String(255))
#     elementid = Column(String(40), nullable=False)
#     elementtype = Column(String(50), nullable=False)
#     strcontent = Column(Text)
#     bincontent1 = Column(LargeBinary)
#     bincontent2 = Column(LargeBinary)


# class TStatustype(Base):
#     __tablename__ = 't_statustypes'

#     status = Column(String(50), primary_key=True)
#     description = Column(String(50))


# class TStereotype(Base):
#     __tablename__ = 't_stereotypes'

#     stereotype = Column(String(255), primary_key=True, nullable=False, index=True)
#     appliesto = Column(String(255), primary_key=True, nullable=False)
#     description = Column(String(255))
#     mfenabled = Column(Integer, server_default=text("0"))
#     mfpath = Column(String(255))
#     metafile = Column(LargeBinary)
#     style = Column(Text)
#     ea_guid = Column(String(50), index=True)
#     visualtype = Column(String(100))


# class TTaggedvalue(Base):
#     __tablename__ = 't_taggedvalue'

#     propertyid = Column(String(40), primary_key=True)
#     elementid = Column(String(40), nullable=False, index=True)
#     baseclass = Column(String(100), nullable=False)
#     tagvalue = Column(Text)
#     notes = Column(Text)


# class TTask(Base):
#     __tablename__ = 't_tasks'

#     taskid = Column(Integer, primary_key=True, server_default=text("nextval(('taskid_seq'::text)::regclass)"))
#     name = Column(String(255))
#     tasktype = Column(String(255))
#     notes = Column(Text)
#     priority = Column(String(255))
#     status = Column(String(255))
#     owner = Column(String(255))
#     startdate = Column(Date)
#     enddate = Column(Date)
#     phase = Column(String(50))
#     history = Column(Text)
#     percent = Column(Integer)
#     totaltime = Column(Integer)
#     actualtime = Column(Integer)
#     assignedto = Column(String(100))


# class TTcf(Base):
#     __tablename__ = 't_tcf'

#     tcfid = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     weight = Column(Float, index=True, server_default=text("1"))
#     value = Column(Float, server_default=text("0"))
#     notes = Column(String(255))


# class TTemplate(Base):
#     __tablename__ = 't_template'

#     templateid = Column(String(40), primary_key=True)
#     templatetype = Column(String(50), nullable=False)
#     templatename = Column(String(100), nullable=False)
#     notes = Column(String(255))
#     style = Column(String(255))
#     template = Column(Text)


# class TTestclas(Base):
#     __tablename__ = 't_testclass'

#     testclass = Column(String(50), primary_key=True)
#     description = Column(String(50))


# class TTestplan(Base):
#     __tablename__ = 't_testplans'

#     planid = Column(String(50), primary_key=True)
#     category = Column(String(100))
#     name = Column(String(150), nullable=False)
#     author = Column(String(255))
#     notes = Column(Text)
#     testplan = Column(Text, nullable=False)


# class TTesttype(Base):
#     __tablename__ = 't_testtypes'

#     testtype = Column(String(12), primary_key=True)
#     description = Column(String(50))
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(String(255))


# class TTrxtype(Base):
#     __tablename__ = 't_trxtypes'

#     description = Column(String(50), index=True)
#     numericweight = Column(Float, index=True, server_default=text("1"))
#     notes = Column(Text)
#     trx = Column(String(255), index=True)
#     trx_id = Column(Integer, primary_key=True, server_default=text("nextval(('trx_id_seq'::text)::regclass)"))
#     style = Column(Text)


# class TUmlpattern(Base):
#     __tablename__ = 't_umlpattern'

#     patternid = Column(Integer, primary_key=True, server_default=text("nextval(('patternid_seq'::text)::regclass)"))
#     patterncategory = Column(String(100))
#     patternname = Column(String(150))
#     style = Column(String(255))
#     notes = Column(Text)
#     patternxml = Column(Text)
#     version = Column(String(50))


# class TVersion(Base):
#     __tablename__ = 't_version'

#     elementid = Column(String(50), primary_key=True, nullable=False, index=True)
#     versionid = Column(String(255), primary_key=True, nullable=False)
#     elementtype = Column(String(100), nullable=False)
#     flags = Column(String(255))
#     externalfile = Column(String(255))
#     notes = Column(String(255))
#     owner = Column(String(255))
#     versiondate = Column(DateTime)
#     branch = Column(String(255))
#     elementxml = Column(Text)


# class TXref(Base):
#     __tablename__ = 't_xref'
#     __table_args__ = (
#         Index('ix_xref_nametype', 'name', 'type'),
#     )

#     xrefid = Column(String(255), primary_key=True)
#     name = Column(String(255), index=True)
#     type = Column(String(255), index=True)
#     visibility = Column(String(255))
#     namespace = Column(String(255))
#     requirement = Column(String(255))
#     Constraint = Column(String(255))
#     behavior = Column(String(255))
#     partition = Column(String(255))
#     description = Column(Text)
#     client = Column(String(255), index=True)
#     supplier = Column(String(255), index=True)
#     link = Column(String(255))


# class TXrefsystem(Base):
#     __tablename__ = 't_xrefsystem'

#     xrefid = Column(String(255), primary_key=True)
#     name = Column(String(255))
#     type = Column(String(255), index=True)
#     visibility = Column(String(255))
#     namespace = Column(String(255))
#     requirement = Column(String(255))
#     Constraint = Column(String(255))
#     behavior = Column(String(255))
#     partition = Column(String(255))
#     description = Column(Text)
#     client = Column(String(255), index=True)
#     supplier = Column(String(255), index=True)
#     link = Column(String(255))
#     toolid = Column(String(50))


# class TXrefuser(Base):
#     __tablename__ = 't_xrefuser'

#     xrefid = Column(String(255), primary_key=True)
#     name = Column(String(255))
#     type = Column(String(255), index=True)
#     namespace = Column(String(255))
#     requirement = Column(String(255))
#     Constraint = Column(String(255))
#     behavior = Column(String(255))
#     partition = Column(String(255))
#     description = Column(Text)
#     client = Column(String(255), index=True)
#     supplier = Column(String(255), index=True)
#     link = Column(String(255))
#     toolid = Column(String(50))
#     visibility = Column(String(255))


# class UsysSchema(Base):
#     __tablename__ = 'usys_schema'

#     V1558 = Column(String(255), primary_key=True)


# class UsysSystem(Base):
#     __tablename__ = 'usys_system'

#     property = Column(String(50), primary_key=True)
#     value = Column(String(50))


# t_usysoldtables = Table(
#     'usysoldtables', metadata,
#     Column('tablename', String(50)),
#     Column('newname', String(50)),
#     Column('relorder', Integer, server_default=text("0")),
#     Column('fixcode', Integer, nullable=False, server_default=text("0"))
# )


# t_usysqueries = Table(
#     'usysqueries', metadata,
#     Column('queryname', String(50)),
#     Column('newname', String(50)),
#     Column('fixcode', Integer, nullable=False)
# )


# class Usystable(Base):
#     __tablename__ = 'usystables'

#     tablename = Column(String(50), primary_key=True)
#     relorder = Column(Integer, server_default=text("0"))
#     displayname = Column(String(50))
#     fromver = Column(String(50))
#     tover = Column(String(50))



from typing import List, Tuple, Optional
from lxml import etree
import logging

from mdg.config import settings
from mdg.uml import (
    UMLAssociation,
    UMLAssociationType,
    UMLAttribute,
    UMLClass,
    UMLEnumeration,
    UMLInstance,
    UMLPackage,
)


logger = logging.getLogger(__name__)


def get_label_name(element):
    label = element.get("label")
    label = label.strip("<div>").strip("</div>")
    return label


def find_label_name(element, name):
    objects = element.findall('object')

    for object in objects:
        label = get_label_name(object)
        if name in label:
            return object
    return None


def parse_uml() -> Tuple[UMLPackage, List[UMLInstance]]:
    test_cases: List[UMLInstance] = []

    # Parse into etree and grab root package
    root = etree.parse(settings['source']).getroot()
    model = root.find('./diagram/mxGraphModel/root')
    if model is None:
        raise ValueError("Cannot find model in XML at path ./diagram/mxGraphModel/root")

    element = find_label_name(model, settings['root_package'])
    if element is None:
        raise ValueError("Root package element not found. Settings has:{}".format(settings['root_package']))

    model_element = find_label_name(model, settings['model_package'])
    if model_element is None:
        raise ValueError("Model element not found. Settings has:{}".format(settings['model_package']))

    model_package: UMLPackage = package_parse(model_element, root, None)

    enumeration_link(model_package)

    return model_package, test_cases


def package_parse(element, root_element, parent_package: Optional[UMLPackage]) -> UMLPackage:
    """ Extract package details, call class parser for classes and self parser for sub-packages.
    Association linking is not done here, but in a 2nd pass using the parse_associations function.
    """

    name = get_label_name(element)
    id = element.get('id')

    package = UMLPackage(id, name, parent_package)
    package._root_element = root_element

    model = root_element.find('./diagram/mxGraphModel/root')
    child_elements = model.findall('object/mxCell[@parent="{}"]'.format(id))

    logger.debug("Added UMLPackage {}".format(package.path))

    # Parse classes
    for element in child_elements:
        object_element = element.find("..")
        element_type = object_element.get("UMLType")

        if element_type == "Class":
            cls = class_parse(package, object_element, root_element)
            package.classes.append(cls)

        elif element_type == "Enumeration":
            enum = enumeration_parse(package, object_element, root_element)
            package.enumerations.append(enum)

        elif element_type == "Package":
            pkg = package_parse(object_element, package._root_element, package)
            package.children.append(pkg)

    # Classes are needed to parse generalisations and associations
    for element in child_elements:
        object_element = element.find("..")
        element_type = object_element.get("UMLType")

        if element_type == "Generalization":
            generalization_parse(package, object_element, root_element)
        elif element_type == "Association":
            association_parse(package, object_element, root_element)
        elif element_type == "Composition":
            association_parse(package, object_element, root_element)

    return package


def find_enumeration_by_name(package, name):
    for enum in package.enumerations:
        if enum.name == name:
            return enum
    for child in package.children:
        res = find_enumeration_by_name(child, name)
        if res is not None:
            return res


def enumeration_link(package):
    root_package = package
    while root_package.parent is not None:
        root_package = root_package.parent

    for cls in package.classes:
        for attr in cls.attributes:
            enum = find_enumeration_by_name(root_package, attr.dest_type)
            if enum is not None:
                attr.classification = enum
    for child in package.children:
        enumeration_link(child)


def generalization_parse(package: UMLPackage, element, root):
    cell = element.find("mxCell")
    source: UMLClass = package.find_by_id(cell.get("source"))
    target: UMLClass = package.find_by_id(cell.get("target"))
    source.generalization = target


def association_parse(package: UMLPackage, element, root):
    id = element.get("id")
    cell = element.find("mxCell")
    source: UMLClass = package.find_by_id(cell.get("source"))
    target: UMLClass = package.find_by_id(cell.get("target"))
    element_type = element.get("UMLType")

    association = UMLAssociation(package, source, target, id, UMLAssociationType[element_type.upper()])

    # Extract multiplicities
    ret = root.findall('.//object[@UMLType="DestinationMultiplicity"]')
    for dest in ret:
        dm = dest.find('mxCell[@parent="{}"]'.format(association.id))
        if dm is not None:
            label = dest.get("label").strip('<div>').strip('</div>')
            association.destination_multiplicity = association.string_to_multiplicity(label)
            break
    ret = root.findall('.//object[@UMLType="SourceMultiplicity"]')
    for dest in ret:
        dm = dest.find('mxCell[@parent="{}"]'.format(association.id))
        if dm is not None:
            label = dest.get("label").strip('<div>').strip('</div>')
            association.source_multiplicity = association.string_to_multiplicity(label)
            break

    # Set association destination name
    destination_name = element.get("destination_name")
    if destination_name is not None:
        association.destination_name = destination_name
    else:
        association.destination_name = target.name
        if association.source_multiplicity[1] == '*' and association.association_type == UMLAssociationType.ASSOCIATION:
            association.destination_name += 's'
        elif association.destination_multiplicity[1] == '*' and association.association_type == UMLAssociationType.COMPOSITION:
            association.destination_name += 's'

    # Set association source name
    source_name = element.get("source_name")
    if source_name is not None:
        association.source_name = source_name
    else:
        association.source_name = source.name
        if association.destination_multiplicity[1] == '*' and association.association_type == UMLAssociationType.ASSOCIATION:
            association.source_name += 's'
        if association.source_multiplicity[1] == '*' and association.association_type == UMLAssociationType.COMPOSITION:
            association.source_name += 's'


def class_parse(package: UMLPackage, element, root) -> UMLClass:
    stereotypes = []
    label = element.get("label").split("<div>")
    if len(label) == 1:
        name = label[0].replace("<b>", "").replace("</b>", "").replace("<i>", "").replace("</i>", "").replace('<br>', "").strip()
    else:
        name = label[-1].replace("</div>", "").replace("<b>", "").replace("</b>", "").replace("<i>", "").replace("</i>", "").replace('<br>', "").strip()
        stereotypes = label[-2].split('&lt;&lt;')[-1].split('&gt;&gt;')[0].split(',')

    id = element.get("id")
    cls = UMLClass(package, name, id)
    for stereotype in stereotypes:
        cls.stereotypes.append(stereotype.strip())

    abstract = element.get("Abstract")
    if abstract is not None and abstract == "True":
        cls.is_abstract = True

    logger.debug("Added UMLClass {}".format(cls.name))

    children = root.findall('./diagram/mxGraphModel/root/mxCell[@parent="{}"]'.format(id))
    # Grab a list of the attribute stereotypes and their heights
    stereotypes = []
    for child in children:
        value = child.get("value")
        if "<<" in value:
            geometry = child.find("mxGeometry")
            stereotypes.append((value.strip("<<").strip(">>"), int(geometry.get("y"))))

    # Create all the attributes
    for child in children:
        value = child.get("value")
        if value[0:2] != "<<":
            attr = attr_parse(cls, child, root, stereotypes)
            cls.attributes.append(attr)

    return cls


def enumeration_parse(package: UMLPackage, element, root) -> UMLEnumeration:
    label = element.get("label").split(",")[-1].split("div>")
    if len(label) == 1:
        name = label[0].strip("<b>").strip("</b>").strip("i>").strip("</i")
    else:
        name = label[-2].strip("<b>").strip("</b></").strip("i>").strip("</i")
    id = element.get("id")
    enum = UMLEnumeration(package, name, id)

    children = root.findall('./diagram/mxGraphModel/root/mxCell[@parent="{}"]'.format(id))
    for child in children:
        enum.values.append(child.get('value'))

    logger.debug("Added UMLEnumeration {}".format(enum.name))

    return enum


def attr_parse(parent: UMLClass, element, root, stereotypes) -> UMLAttribute:
    value = element.get("value").strip("<div>").strip("</div>").strip("<br")
    # height = int(element.find("mxGeometry").get("y"))

    dq = []
    if "{dq_even}" in value:
        dq.append('even')
        value = value.replace("{dq_even}", "").strip()

    is_id = False
    if "{id}" in value:
        is_id = True
        value = value.replace("{id}", "").strip()

    visibility: bool = False
    if value.startswith("+"):
        visibility = True
    value = value.strip("+").strip("-").strip()

    name, attr_type = value.split(":")

    attr = UMLAttribute(parent, name, element.get('id'))
    if is_id:
        attr.is_id = is_id
        parent.id_attribute = attr
    attr.visibility = visibility

    attr.validations = dq

    attr.set_type(attr_type.strip())

    return attr



# import re
from typing import List, Tuple, Union
from lxml import etree
import logging
# from decimal import Decimal

from mdg.config import settings
from mdg.uml import (
    UMLAssociation,
    UMLAttribute,
    UMLClass,
    UMLComponent,
    UMLEnumeration,
    UMLInstance,
    UMLPackage,
    # UMLAssociationType,
)


logger = logging.getLogger(__name__)


def parse_uml() -> Tuple[UMLPackage, List[UMLInstance]]:
    """ Root package parser entry point.
    """
    test_cases: List[UMLInstance] = []

    # Parse into etree and grab root package
    logger.debug(f"loading etree from {settings['source']}")
    tree = etree.parse(settings['source'])
    root_package = tree.find('XMI.content')

    # Find the element that is the root for models
    logger.info("Parsing models")
    model_element = root_package.xpath("//Model_Management.Package[@xmi.id='%s']" % settings['model_package'])
    if len(model_element) == 0:
        raise ValueError("Model package element not found. Settings has:{}".format(settings['model_package']))
    model_element = model_element[0]

    # Create our root model UMLPackage and parse in 3 passes
    model_package = package_parse(model_element, root_package, None)
    # logger.debug("Parsing inheritance")
    # model_package_parse_inheritance(model_package)
    logger.debug("Parsing associations")
    package_parse_associations(model_package, model_element)
    # logger.debug("Sorting objects")
    # package_sort_classes(model_package)

    # if 'test_package' in settings.keys():
    #     logger.info("Parsing test cases")

    #     # Find the element that is the root for test data
    #     test_element = element.xpath("//packagedElement[@name='%s']" % settings['test_package'], namespaces=ns)
    #     if len(test_element) == 0:
    #         raise ValueError("Test packaged element not found. Settings has:{}".format(settings['test_package']))
    #     test_element = test_element[0]

    #     # Create our root test data UMLPackage and parse in 2 passes. Does not support inheritance
    #     e_type = test_element.get('{%s}type' % ns['xmi'])
    #     if e_type == 'uml:Package':
    #         test_package = package_parse(test_element, tree, None)
    #         package_parse_associations(test_package, test_element, test_element)
    #         logger.debug("Parsing inheritance")
    #         test_package_parse_inheritance(test_package, model_package)

    #         # With our test package parsed, we must return a list of instances instead of hierarchy of packages
    #         test_cases = parse_test_cases(test_package)
    return model_package, test_cases


def parse_test_cases(package) -> List[UMLInstance]:
    """ Looks through package hierarchy for instances with request or response stereotype
    and returns list of instances.
    :rtype: list<UMLInstance>
    """
    test_cases = []

    for instance in package.instances:
        if instance.stereotype in ['request', 'response']:
            test_cases.append(instance)

    for child in package.children:
        res = parse_test_cases(child)
        if res:
            test_cases += res

    return test_cases


def package_parse(element, root_element, parent_package):
    """ Extract package details, call class parser for classes and self parser for sub-packages.
    Associations are not done here, but in a 2nd pass using the parse_associations function.
    :param element:
    :param root_element:
    :return:
    :rtype: UMLPackage
    """
    id = element.get('xmi.id')
    name = element.find('Foundation.Core.ModelElement.name').text
    package = UMLPackage(id, name, parent_package)
    package._element = element
    package._root_element = root_element

    # package.documentation = properties.get('documentation')
    # if package.documentation is None:
    #     package.documentation = ""

    # project = detail.find('project')
    # package.status = project.get('status')

    # diagram_elements = root_element.xpath("//diagrams/diagram/model[@package='%s']" % package.id)
    # for diagram_model in diagram_elements:
    #     diagram = diagram_model.getparent()
    #     package.diagrams.append(diagram.get('{%s}id' % ns['xmi']))

    logger.debug("Added UMLPackage {}".format(package.path))
    owned = element.find('Foundation.Core.Namespace.ownedElement')
    package_parse_children(owned, package)
    return package


def package_parse_children(element, package):

    # Loop through all child elements and create nodes for classes and sub packages
    for child in element:
        e_type = child.tag

        if e_type == 'Model_Management.Package':
            pkg = package_parse(child, package._root_element, package)
            package.children.append(pkg)

        elif e_type == 'Foundation.Core.Class':
            cls = class_parse(package, child)
            if cls.name is not None:
                package.classes.append(cls)

#         elif e_type == 'uml:Enumeration':
#             enumeration = enumeration_parse(package, child)
#             if enumeration.name is not None:
#                 package.enumerations.append(enumeration)
#             # print(f"Enum: {package.path}{enumeration}")


def package_parse_associations(package, element):
    """ Packages and classes should already have been parsed so now we link classes for each association.
    This gets messy as XMI output varies based on association type.
    This supports both un-specified and source to destination directional associations
    :param root_element:
    :param element: XML Element
    :type package: UMLPackage
    """
    owned = element.find('Foundation.Core.Namespace.ownedElement')
    for child in owned:
        e_type = child.tag

        if e_type == 'uml:Package':
            pkg = package_parse_associations(package, child)
            package.children.append(pkg)

        if e_type == 'Foundation.Core.Association':
            association = association_parse(package, child)
            package.associations.append(association)


def association_parse(package, element):
    id = element.get('xmi.id')

    ends = element.find('Foundation.Core.Association.connection')
    if len(ends) != 2:
        logger.warn("Two ends of association not provided with id {id}")
        return None
    source_id = ends[0].find("Foundation.Core.AssociationEnd.type")[0].get('xmi.idref')
    dest_id = ends[1].find("Foundation.Core.AssociationEnd.type")[0].get('xmi.idref')

    source = package.root_package.find_by_id(source_id)
    dest = package.root_package.find_by_id(dest_id)

    association = UMLAssociation(package, source, dest, id)

    # Extract multiplicity for source
    multi = ends[0].find('Foundation.Core.AssociationEnd.multiplicity')[0][0][0]
    source_lower = multi.find('Foundation.Data_Types.MultiplicityRange.lower').text
    if source_lower == '-1':
        source_lower = '*'
    source_upper = multi.find('Foundation.Data_Types.MultiplicityRange.upper').text
    if source_upper == '-1':
        source_upper = '*'
    association.source_multiplicity = (source_lower, source_upper)

    # Extract multiplicity for dest
    multi = ends[1].find('Foundation.Core.AssociationEnd.multiplicity')[0][0][0]
    dest_lower = multi.find('Foundation.Data_Types.MultiplicityRange.lower').text
    if dest_lower == '-1':
        dest_lower = '*'
    dest_upper = multi.find('Foundation.Data_Types.MultiplicityRange.upper').text
    if dest_upper == '-1':
        dest_upper = '*'
    association.destination_multiplicity = (dest_lower, dest_upper)

    print(f"{source_id} to {dest_id} | {source_lower} | {source_upper} | {dest_lower} | {dest_upper} | {association.cardinality}")

#     assoc_type = source_element.get("aggregation")
#     if assoc_type == "composite":
#         association.association_type = UMLAssociationType.COMPOSITION
#         dest.composed_of.append(source)
#     elif assoc_type == "shared":
#         association.association_type = UMLAssociationType.AGGREGATION

#     # If it's an association to or from a multiple then pluralize the name
#     # TODO: Allow pluralized name to be specified in UML
#     # Use opposing ends class name as attribute name for association
    association.destination_name = association.destination.name.lower()
    if association.destination_multiplicity[1] == '*':
        association.destination_name += 's'
    association.source_name = association.source.name.lower()
    if association.source_multiplicity[1] == '*':
        association.source_name += 's'

#     # Allow explicit naming
#     if dest_element.get('name') is not None:
#         association.source_name = dest_element.get('name')
#     if source_element.get('name') is not None:
#         association.destination_name = source_element.get('name')

    return association


# def enumeration_parse(package, element):
#     enumeration = UMLEnumeration(package, element.get('name'), element.get('{%s}id' % ns['xmi']))

#     # Loop through class elements children for values.
#     for child in element:
#         e_type = child.get('{%s}type' % ns['xmi'])
#         if e_type == 'uml:EnumerationLiteral':
#             enumeration.values.append(child.get('name'))
#     logger.debug(f"Added UMLEnumeration {enumeration.name}")
#     return enumeration


def class_parse(package, element):
    id = element.get('xmi.id')
    name = element.find('Foundation.Core.ModelElement.name').text
    cls: UMLClass = UMLClass(package, name, id)
    if element.find('Foundation.Core.GeneralizableElement.isAbstract').get('value') == 'true':
        cls.is_abstract = True
    else:
        cls.is_abstract = False

    print(f"Added UMLClass {cls.id}: {cls.name}")

    # Loop through class elements children for attributes.
    features = element.find('Foundation.Core.Classifier.feature')
    if features is not None:
        for child in features:
            e_type = child.tag

            if e_type == 'Foundation.Core.Attribute':
                attr = attr_parse(cls, child)
                cls.attributes.append(attr)

    return cls


def attr_parse(parent: Union[UMLClass, UMLEnumeration, UMLComponent], element):
    id = element.get('xmi.id')
    name = element.find('Foundation.Core.ModelElement.name').text
    attr = UMLAttribute(parent, name, id)

    attr.visibility = element.find('Foundation.Core.ModelElement.visibility').get('xmi.value')
#     type_elem = element.find('type')
#     if type_elem is not None:
#         type_id = type_elem.get('{%s}idref' % ns['xmi'])
#         if type_id[:4] == 'EAID':
#             attr.classification_id = type_id
#     else:
#         logging.error(f"Attribute {attr.name} of class {parent} does not have a type")

#     alias_node = detail.find('style')
#     attr.alias = alias_node.get('value')

#     xrefs = detail.find('xrefs')
#     if xrefs.get('value') is not None and 'NAME=isID' in xrefs.get('value'):
#         attr.is_id = True
#         attr.parent.id_attribute = attr
#     else:
#         attr.is_id = False

    print(f"Added attr {attr.id}: {attr.name}")
    return attr



import time
import logging
import sqlalchemy
from sqlalchemy.orm import Session

from ..parse.sparx_db import TPackage
from ..generate import generate

from mdg.config import settings

logger = logging.getLogger(__name__)


def poller(duration):
    engine = sqlalchemy.create_engine(f"{settings['source']}", echo=False, future=True, isolation_level="READ UNCOMMITTED")
    with Session(engine) as session:
        done = False
        version = None

        # Find the package with the model nodes. Can specify either EA GUID or name
        # If guid make sure value in recipie is quoted - model_package: "{D6D3BF36-E897-4a8b-8CA9-62ADAAD696ED}"
        if settings['model_package'][0] == "{":
            stmt = sqlalchemy.select(TPackage).where(TPackage.ea_guid == settings['model_package'])
        else:
            stmt = sqlalchemy.select(TPackage).where(TPackage.name == settings['model_package'])

        while not done:
            model_tpackage: TPackage = session.execute(stmt).scalars().first()
            session.commit()
            if model_tpackage is None:
                raise ValueError("Model package element not found. Settings has:{}".format(settings['model_package']))

            if version is not None and model_tpackage.version != version:
                logger.info(f"Detected version change in model package. Version {model_tpackage.version}")
                generate()

            version = model_tpackage.version
            time.sleep(duration)



#!/usr/bin/python
from mdg.parse import parse


class ClassValidationError(object):
    def __init__(self, package, cls, error):
        self.package = package
        self.error = error
        self.cls = cls

    def __repr__(self):
        return "Class error: {}{} | {}".format(self.package.path, self.cls.name, self.error)


class AttributeValidationError(object):
    def __init__(self, package, cls, attr, error):
        self.package = package
        self.error = error
        self.cls = cls
        self.attr = attr

    def __repr__(self):
        return "Attribute error: {}{}.{} | {}".format(self.package.path, self.cls.name, self.attr.name, self.error)


def validate_package(package):
    errors = []

    for cls in package.classes:
        if cls.id_attribute is None and cls.generalization is None and not cls.is_abstract:
            errors.append(ClassValidationError(package, cls, "no primary key"))
        elif cls.generalization is not None:
            if cls.generalization.id_attribute is not None and cls.generalization.id_attribute != cls.id_attribute:
                errors.append(ClassValidationError(package, cls, "primary key in both class and generalization class"))

        for attr in cls.attributes:
            if attr.stereotype == "auto" and attr.type not in ("int", "bigint"):
                errors.append(AttributeValidationError(package, cls, attr, "auto increment field must be int or bigint"))

    for child in package.children:
        errors += validate_package(child)

    return errors


def validate():
    """ Loads XMI file from settings as an etree
        Calls XMI parser to turn model and tests into python native (see UML metamodel)
        Calls output functions to render for model and tests
    """

    model_package, test_cases = parse()

    errors = validate_package(model_package)
    if len(errors) > 0:
        print("Validation Errors:")
        for error in errors:
            print("    {}".format(error))

    return errors



import unittest

from mdg.uml import UMLAttribute, UMLClass, UMLPackage, UMLAssociation
from mdg.uml import dumps as uml_dumps


class TestDumps(unittest.TestCase):
    def setUp(self):
        self.package = UMLPackage("id", "name")

        cls1 = UMLClass(self.package, "test1", 1)
        attr = UMLAttribute(cls1, 'Test Att', 2)
        cls1.attributes.append(attr)
        attr = UMLAttribute(cls1, 'Test Att 2', 3)
        cls1.attributes.append(attr)

        cls2 = UMLClass(self.package, "test2", 4)
        attr = UMLAttribute(cls2, 'Test Att 3', 5)
        cls2.attributes.append(attr)
        attr = UMLAttribute(cls2, 'Test Att 4', 6)
        cls2.attributes.append(attr)

        self.package.classes.append(cls1)
        self.package.classes.append(cls2)

        assoc = UMLAssociation(self.package, cls1, cls2, 7)
        self.package.associations.append(assoc)

    def test_object_serialise(self):
        data: str = uml_dumps(self.package)
        self.assertGreater(data.find('"associations": [{'), -1)



from django.shortcuts import render


def index(request):
    return render(request, 'index.html', {
        'foo': 'bar',
    })


def streams(request):
    return render(request, 'streams.html', {
        'foo': 'bar',
    })



# import enum
from typing import List, Tuple, Optional, Union
import logging
import re
from decimal import Decimal

import sqlalchemy
from sqlalchemy.orm import Session

from mdg.parse.sparx_db_models import (
    TObject,
    TPackage,
    TXref,
    TAttribute,
    TAttributeconstraint,
    TConnector,
)

from mdg.config import settings
from mdg.uml import (
    UMLAssociation,
    UMLAttribute,
    UMLClass,
    UMLEnumeration,
    UMLInstance,
    UMLPackage,
    UMLAssociationType,
    SearchTypes,
)


logger = logging.getLogger(__name__)


def parse_uml() -> Tuple[UMLPackage, List[UMLInstance]]:
    """ Root package parser entry point.
    """
    test_cases: List[UMLInstance] = []

    engine = sqlalchemy.create_engine(f"{settings['source']}", echo=False, future=True)
    with Session(engine) as session:

        # Find the package with the model nodes. Can specify either EA GUID or name
        # If guid make sure value in recipie is quoted - model_package: "{D6D3BF36-E897-4a8b-8CA9-62ADAAD696ED}"
        if settings['model_package'][0] == "{":
            stmt = sqlalchemy.select(TPackage).where(TPackage.ea_guid == settings['model_package'])
        else:
            stmt = sqlalchemy.select(TPackage).where(TPackage.name == settings['model_package'])
        model_tpackage: TPackage = session.execute(stmt).scalars().first()
        if model_tpackage is None:
            raise ValueError("Model package element not found. Settings has:{}".format(settings['model_package']))
        logger.debug(f"Model Object: {model_tpackage.package_id}: {model_tpackage.name}")

        # Create our root model UMLPackage and parse in 3 passes
        model_package = package_parse(session, model_tpackage, None)
        logger.debug("Parsing associations and inheritance")
        package_parse_associations(session, model_package)
        logger.debug("Sorting objects")
        package_sort_classes(model_package)

    if 'test_package' in settings.keys():
        logger.info("Parsing test cases")
        if settings['test_package'][0] == "{":
            stmt = sqlalchemy.select(TPackage).where(TPackage.ea_guid == settings['test_package'])
        else:
            stmt = sqlalchemy.select(TPackage).where(TPackage.name == settings['test_package'])
        test_tpackage: TPackage = session.execute(stmt).scalars().first()
        if test_tpackage is None:
            raise ValueError("Test package element not found. Settings has:{}".format(settings['test_package']))
        test_package = package_parse(session, test_tpackage, None)
        package_parse_associations(session, test_package)
        test_package_parse_inheritance(test_package, model_package)
        test_cases = parse_test_cases(test_package)

    return model_package, test_cases


def parse_test_cases(package: UMLPackage) -> List[UMLInstance]:
    """ Looks through package hierarchy for instances with request or response stereotype
    and returns list of instances.
    """
    test_cases = []

    for instance in package.instances:
        if instance.stereotype in ['request', 'response']:
            test_cases.append(instance)

    for child in package.children:
        res = parse_test_cases(child)
        if res:
            test_cases += res

    return test_cases


def package_parse(session, tpackage: TPackage, parent_package: Optional[UMLPackage]):
    """ Extract package details, call class parser for classes and self parser for sub-packages.
    Associations are not done here, but in a 2nd pass using the parse_associations function.
    :param element:
    :param root_element:
    :return:
    :rtype: UMLPackage
    """

    package = UMLPackage(tpackage.package_id, tpackage.name, parent_package)

    stmt = sqlalchemy.select(TObject).where(TObject.ea_guid == tpackage.ea_guid)
    tobject = session.execute(stmt).scalars().first()

    package.stereotype = f"{tobject.stereotype}"
    if package.stereotype is not None:
        package.inherited_stereotypes.append((package.stereotype, package))

    package.documentation = tobject.note
    if package.documentation is None:
        package.documentation = ""

    package.status = tobject.status
    package.version = f"{tpackage.version}"

    logger.debug("Added UMLPackage {}".format(package.path))

    package_parse_children(session, package)
    return package


def package_parse_children(session, package: UMLPackage):
    stmt = sqlalchemy.select(TObject).where(TObject.package_id == package.id)

    # Loop through all child elements and create nodes for classes and sub packages
    for child_tobject in session.execute(stmt).scalars().all():
        if child_tobject.object_type == 'Package':
            stmt = sqlalchemy.select(TPackage).where(TPackage.ea_guid == child_tobject.ea_guid)
            tpackage = session.execute(stmt).scalars().first()
            pkg = package_parse(session, tpackage, package)
            package.children.append(pkg)

        elif child_tobject.object_type == 'Class':
            cls = class_parse(session, package, child_tobject)
            if cls.name is not None:
                package.classes.append(cls)

        elif child_tobject.object_type == 'Object':
            ins = instance_parse(session, package, child_tobject)
            if ins.name is not None:
                package.instances.append(ins)

        elif child_tobject.object_type == 'Enumeration':
            enumeration = enumeration_parse(session, package, child_tobject)
            package.enumerations.append(enumeration)


def package_parse_associations(session, package: UMLPackage):
    """ Packages and classes should already have been parsed so now we link classes for each association.
    """
    for cls in package.classes:
        stmt = sqlalchemy.select(TConnector).where(TConnector.start_object_id == cls.id)

        for connector in session.execute(stmt).scalars().all():
            dest = package.root_package.find_by_id(connector.end_object_id, SearchTypes.CLASS)
            if dest is None:
                logger.warn(f"Cannot find associated class from {cls}. Association Id:{connector.id}, Destination Id: {connector.end_object_id}")

            elif connector.connector_type in ["Association", "Aggregation"]:
                association = association_parse(session, connector, package, cls, dest)
                if association is not None:
                    package.associations.append(association)

            elif connector.connector_type == "Generalization":
                cls.generalization = dest
                dest.specialized_by.append(cls)
                if cls.id_attribute is None:
                    cls.id_attribute = dest.id_attribute

        # Find enumeration attributes of class and link to attribute
        for attr in cls.attributes:
            if attr.classification_id is not None:
                attr.classification = package.root_package.find_by_id(attr.classification_id, SearchTypes.ENUM)
                if attr.classification is None:
                    logger.warn("Cannot find expected classification for {} of attribute {}. Id={}".format(attr.dest_type, attr.name, attr.classification_id))

    for ins in package.instances:
        stmt = sqlalchemy.select(TConnector).where(TConnector.start_object_id == ins.id)

        for connector in session.execute(stmt).scalars().all():
            dest = package.root_package.find_by_id(connector.end_object_id, SearchTypes.INSTANCE)

            association = association_parse(session, connector, package, ins, dest)
            if association is not None:
                package.associations.append(association)

    for package_child in package.children:
        package_parse_associations(session, package_child)


def test_package_parse_inheritance(test_package, model_package):
    """ Links instances with the class they are instances of
        and sets the attribute types which wern't in the run_state where instance attrs are created from
    """

    for ins in test_package.instances:
        if ins.classification_id is not None:
            ins.classification = model_package.find_by_id(ins.classification_id)
            if ins.classification is None:
                ins.classification = test_package.find_by_id(ins.classification_id)
                if ins.classification is None:
                    logger.warn(f"Cannot find class which instance named {ins.name} is from id={ins.classification_id}")

            if ins.classification is not None:
                for attr in ins.attributes:
                    for cls_attr in ins.classification.attributes:
                        if attr.name == cls_attr.name:
                            attr.type = cls_attr.type
                            if attr.type.lower() in ['int', 'integer']:
                                attr.value = int(attr.value)
                            elif attr.type.lower() == ['float']:
                                attr.value = float(attr.value)
                            elif attr.type.lower() == ['decimal']:
                                attr.value = Decimal(attr.value)
                            break
        else:
            logger.warn("Instance object which is not from any class: id={}".format(ins.id))

    for child in test_package.children:
        test_package_parse_inheritance(child, model_package)


def package_sort_classes(package):
    # print(f"SORT {package.name}")
    unordered_list = package.classes
    unordered_composed_list = []
    for cls in unordered_list:
        if cls.composed_of != []:
            unordered_composed_list.append(cls)
            package.classes.remove(cls)
            # print(f"  REMOVE {cls.name}")

    # This is crap, do some propper sorting
    unordered_list = unordered_composed_list
    composed_list = []
    for cls1 in unordered_list:
        for cls2 in cls1.composed_of:
            if cls2 in unordered_composed_list:
                unordered_composed_list.remove(cls2)
                composed_list.append(cls2)

    composed_list += unordered_composed_list
    package.classes += composed_list

    for child in package.children:
        package_sort_classes(child)


def instance_parse(session, package: UMLPackage, tobject: TObject):
    ins = UMLInstance(package, tobject.name, tobject.object_id)

    ins.stereotype = tobject.stereotype
    ins.documentation = tobject.note
    ins.status = tobject.status

    # We need to link this instance to the class it is an instance of
    ins.classification_id = tobject.classifier

    # Create attributes for each item found in the runstate
    run_state = tobject.runstate
    if run_state is not None:
        vars = run_state.split('@ENDVAR;')
        for var in vars:
            if var != '':
                variable, value = (var.split(';')[1:3])
                attr = UMLAttribute(ins, variable.split('=')[1], value.split('=')[1])
                attr.value = value.split('=')[1]
                ins.attributes.append(attr)
    else:
        logger.info(f"No runstate found for UMLInstance {ins.name} | {ins.id}")
    logger.debug(f"Added UMLInstance {ins.name}")
    return ins


def association_parse(session, tconnector: TConnector, package: UMLPackage, source: Union[UMLClass, UMLInstance], dest: Union[UMLClass, UMLInstance]):
    association = UMLAssociation(package, source, dest, tconnector.connector_id)
    association.documentation = tconnector.notes

    if tconnector.connector_type == "Aggregation":
        if tconnector.subtype == "Strong":
            association.association_type = UMLAssociationType.COMPOSITION
            if isinstance(dest, UMLClass) and isinstance(source, UMLClass):
                dest.composed_of.append(source)
        else:
            association.association_type = UMLAssociationType.AGGREGATION

    association.source_multiplicity = association.string_to_multiplicity(tconnector.sourcecard)
    association.destination_multiplicity = association.string_to_multiplicity(tconnector.destcard)

    # Use opposing ends class name as attribute name for association
    association.source_name = association.source.name.lower()
    association.destination_name = association.destination.name.lower()

    # If it's an association to or from a multiple then pluralize the name
    if association.destination_multiplicity[1] == '*':
        association.destination_name += 's'
    if association.source_multiplicity[1] == '*':
        association.source_name += 's'

    # Allow explicit naming
    if tconnector.sourcerole is not None:
        association.source_name = tconnector.sourcerole
    if tconnector.destrole is not None:
        association.destination_name = tconnector.destrole

    logging.debug(f"Created {association.cardinality.name} {association.association_type.name} from {association.source_name} to {association.destination_name}")
    return association


def enumeration_parse(session, package: UMLPackage, tobject: TObject):
    enumeration = UMLEnumeration(package, tobject.name, tobject.object_id)

    # Loop through class elements children for values.
    stmt = sqlalchemy.select(TAttribute).where(TAttribute.object_id == tobject.object_id)
    for tattribute in session.execute(stmt).scalars().all():
        enumeration.values.append(tattribute.name)

    logging.debug(f"Added UMLEnumeration {enumeration.id}:{enumeration.name}")
    return enumeration


def class_parse(session, package: UMLPackage, tobject: TObject):
    cls: UMLClass = UMLClass(package, tobject.name, tobject.object_id)
    if tobject.abstract == '1':
        cls.is_abstract = True
    else:
        cls.is_abstract = False

    cls.alias = tobject.alias
    cls.status = tobject.status
    cls.phase = tobject.phase
    if tobject.note is not None:
        cls.documentation = tobject.note
    else:
        cls.documentation = ""

    # TXref
    #   description @STEREO;Name=notifiable;GUID={ADC4E914-13DD-4f1b-A9DB-EDCB89896228};@ENDSTEREO;@STEREO;Name=auditable;GUID={C5DA655B-B862-4a27-96F8-FEB0B2EDD529};@ENDSTEREO;
    stmt = sqlalchemy.select(TXref).where(TXref.client == tobject.ea_guid, TXref.name == "Stereotypes")
    txref = session.execute(stmt).scalars().first()
    if txref is not None:
        cls.stereotypes = re.findall('@STEREO;Name=(.*?);', txref.description)

    logger.debug(f"Added UMLClass {cls.package.path}{cls.name} | Stereotypes: {cls.stereotypes}")

    stmt = sqlalchemy.select(TAttribute).where(TAttribute.object_id == tobject.object_id)
    for tattribute in session.execute(stmt).scalars().all():
        attr = attr_parse(session, cls, tattribute)
        cls.attributes.append(attr)

    return cls


def attr_parse(session, parent: UMLClass, tattribute: TAttribute):
    attr = UMLAttribute(parent, tattribute.name, tattribute.id)

    # attr.visibility = element.get('visibility')

    type_elem = tattribute.type
    if type_elem is not None:
        attr.set_type(type_elem)
    else:
        logging.error(f"Attribute {attr.name} of class {parent} does not have a type")

    if tattribute.classifier is not None and tattribute.classifier != "0":
        attr.classification_id = tattribute.classifier

    if tattribute.notes is not None:
        attr.documentation = tattribute.notes
    else:
        attr.documentation = ""

    # @PROP=@NAME=isID@ENDNAME;@TYPE=Boolean@ENDTYPE;@VALU=1@ENDVALU;@PRMT=@ENDPRMT;@ENDPROP;
    stmt = sqlalchemy.select(TXref).where(TXref.client == tattribute.ea_guid, TXref.name == "CustomProperties")
    txref = session.execute(stmt).scalars().first()
    if txref is not None:
        attr.is_id = bool(re.findall('@NAME=isID.*@VALU=1(.*?)@ENDVALU;', txref.description))
        if attr.is_id and isinstance(attr.parent, UMLClass):
            attr.parent.id_attribute = attr
    else:
        attr.is_id = False

    attr.stereotype = tattribute.stereotype

    stmt = sqlalchemy.select(TAttributeconstraint).where(TAttributeconstraint.id == tattribute.id)
    for constraint in session.execute(stmt).scalars().all():
        if constraint.Constraint == 'unique':
            attr.is_unique = True
        elif constraint.Constraint.startswith('length'):
            attr.length = int(constraint.Constraint.split("=")[1])

    logger.debug(f"Added Attribute {attr.parent.name}/{attr.name}")

    return attr



#!/usr/bin/python
from __future__ import annotations
from decimal import Decimal
from typing import List, Union, Optional, Any, Tuple
from enum import Enum
import json

from mdg import generation_fields
from mdg.config import settings
from mdg.tools.io import obj_to_dict


# TODO: Add a UMLObject abstract class which UMLClass, UMLEnumeration and UMLComponent inherit from. Update obj_to_dict to handle inheritance.

class UMLStatuses(Enum):
    Proposed = 1
    Approved = 2
    Implemented = 3
    Mandatory = 4
    Validated = 5


class UMLAssociationType(Enum):
    ASSOCIATION = 1
    COMPOSITION = 2
    AGGREGATION = 3


class Cardinality(Enum):
    MANY_TO_ONE = 1
    ONE_TO_MANY = 2
    MANY_TO_MANY = 3
    ONE_TO_ONE = 4


class SearchTypes(Enum):
    ENUM = "enumeration"
    CLASS = "class"
    PACKAGE = "package"
    INSTANCE = "instance"


class UMLPackage:
    parent: Optional[UMLPackage]
    children: List[UMLPackage]

    classes: List[UMLClass]
    components: List[UMLComponent]
    associations: List[UMLAssociation]
    instances: List[UMLInstance]
    enumerations: List[UMLEnumeration]

    stereotype: Optional[str]
    name: str
    id: Union[int, str]
    diagrams: Any
    documentation: str
    status: Optional[UMLStatuses]
    version: Optional[str]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = ['classes', 'associations', 'children', 'enumerations', 'components']

    def __init__(self, id: Union[int, str], name: str, parent: Optional[UMLPackage] = None, stereotype: Optional[str] = None) -> None:
        self.parent = parent
        self.children = []

        self.classes = []
        self.components = []
        self.associations = []
        self.instances = []
        self.enumerations = []

        self.stereotype = stereotype
        self.inherited_stereotypes: List[Tuple[str, UMLPackage]] = []
        self.name = name
        self.id = id
        self.diagrams = []
        self.documentation = ""
        self.status = None
        self.version = None

        self._element: Any = None
        self._root_element: Any = None
        self.root_package: UMLPackage
        self.path: str

        if self.parent is None:
            self.root_package = self
            self.path = '/' + self.name + '/'
        else:
            self.root_package = self.parent.root_package
            self.inherited_stereotypes += self.parent.inherited_stereotypes
            self.path = self.parent.path + self.name + '/'

    def __str__(self):
        return f"{self.name}"

    def find_by_id(self, id: Union[int, str], find_type: Optional[SearchTypes] = None):
        """ Finds UMLPackage, UMLClass, UMLEnumeration or UMLInstance object with specified Id
        Looks for classes part of this package and all sub-packages
        """
        if str(self.id) == str(id) and find_type in (None, SearchTypes.PACKAGE):
            return self

        for cls in self.classes:
            if str(cls.id) == str(id) and find_type in (None, SearchTypes.CLASS):
                return cls

        for ins in self.instances:
            if str(ins.id) == str(id) and find_type in (None, SearchTypes.INSTANCE):
                return ins

        for enum in self.enumerations:
            if str(enum.id) == str(id) and find_type in (None, SearchTypes.ENUM):
                return enum

        for child in self.children:
            res = child.find_by_id(id, find_type)
            if res is not None:
                return res

    def add_child(self, id: Union[int, str], name: str, stereotype=None) -> UMLPackage:
        package: UMLPackage = UMLPackage(id, name, self, stereotype)
        self.children.append(package)
        return package

    def get_all_classes(self) -> list:
        """ Returns a list of all classes in the tree of packages """

        result = self.classes.copy()
        for package in self.children:
            result += package.get_all_classes()
        return result

    def get_all_enums(self) -> list:
        """ Returns a list of all enums in the tree of packages """

        result = self.enumerations.copy()
        for package in self.children:
            result += package.get_all_enums()
        return result


class UMLInstance:
    package: UMLPackage
    attributes: List[UMLAttribute]
    stereotype: Optional[str]
    name: str
    id: Union[int, str]
    documentation: str
    status: Optional[UMLStatuses]
    associations_from: List[UMLAssociation]
    associations_to: List[UMLAssociation]
    classification: Optional[UMLClass]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = ['attributes', ]

    def __init__(self, package: UMLPackage, name: str, id: Union[int, str]) -> None:
        self.attributes = []
        self.associations_from = []
        self.associations_to = []
        self.package = package
        self.stereotype = None
        self.name = name
        self.id = id
        self.documentation = ""
        self.status = None
        self.classification_id: Union[None, int, str] = None

    def __str__(self) -> str:
        return f"{self.name}"


class UMLAssociation:
    id: Union[int, str]
    documentation: str

    association_type: UMLAssociationType
    source: Union[UMLClass, UMLInstance, UMLComponent]
    source_name: Optional[str]
    source_multiplicity: Tuple[str, str]

    destination: Union[UMLClass, UMLInstance, UMLComponent]
    destination_multiplicity: Tuple[str, str]
    destination_name: Optional[str]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = []

    def __init__(self, package: UMLPackage, source: Union[UMLClass, UMLInstance, UMLComponent], destination: Union[UMLClass, UMLInstance, UMLComponent], id: Union[int, str], assoc_type=UMLAssociationType.ASSOCIATION):
        self.package: UMLPackage = package
        self.id = id
        self.documentation = ""

        self.association_type = assoc_type
        self.source = source
        self.source_name = None
        self.source_multiplicity = ('0', '0')
        source.associations_from.append(self)

        self.destination = destination
        self.destination_multiplicity = ('0', '0')
        self.destination_name = None
        destination.associations_to.append(self)

    def __str__(self) -> str:
        return f"{self.source_name}({self.source_multiplicity}) -> {self.destination_name}({self.destination_multiplicity})"

    @property
    def cardinality(self):
        cardinality = None
        # Use multiplicities to calculate the CARDINALITY of THYE association
        if self.source_multiplicity[1] == '*' and self.destination_multiplicity[1] in ('0', '1'):
            cardinality = Cardinality.MANY_TO_ONE
        elif self.destination_multiplicity[1] == '*' and self.source_multiplicity[1] in ('0', '1'):
            cardinality = Cardinality.ONE_TO_MANY
        elif self.destination_multiplicity[1] == '*' and self.source_multiplicity[1] == '*':
            cardinality = Cardinality.MANY_TO_MANY
        elif self.destination_multiplicity[1] in ('0', '1') and self.source_multiplicity[1] in ('0', '1'):
            cardinality = Cardinality.ONE_TO_ONE

        return cardinality

    def string_to_multiplicity(self, value):
        map = {
            "0..*": ("0", "*"),
            "1..*": ("1", "*"),
            "1": ("0", "1"),
            "0..1": ("0", "1"),
            "1..1": ("1", "1"),
            "*..1": ("*", "1"),
            "*..0": ("*", "0"),
            "*": ("0", "*"),
        }
        output = ('1', '1')
        if value in map.keys():
            output = map[value]

        return output


class UMLEnumeration:
    package: UMLPackage
    name: str
    id: Union[int, str]
    values: List[str]
    documentation: str
    status: Optional[UMLStatuses]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = []

    def __init__(self, package: UMLPackage, name: str, id: Union[int, str]):
        self.values = []
        self.package = package
        self.name = name
        self.id = id
        self.documentation = ""
        self.status = None
        self.id_attribute: Optional[UMLAttribute] = None

    def __str__(self) -> str:
        return f"{self.name}"


class UMLClass:
    package: UMLPackage
    name: str
    alias: Optional[str]
    id: Union[int, str]
    attributes: List[UMLAttribute]
    associations_from: List[UMLAssociation]
    associations_to: List[UMLAssociation]
    generalization: Optional[UMLClass]
    specialized_by: List[UMLClass]
    stereotypes: List[str]
    documentation: str
    is_abstract: bool
    status: Optional[UMLStatuses]
    composed_of: List[UMLClass]
    phase: Optional[Decimal]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = ['attributes', ]

    def __init__(self, package: UMLPackage, name: str, id: Union[int, str]):
        self.name = name
        self.alias = None
        self.id = id
        self.attributes = []
        self.associations_from = []
        self.associations_to = []
        self.package = package
        self.generalization = None
        self.generalization_id: Union[None, int, str] = None
        self.specialized_by = []
        self.stereotypes = []
        self.id_attribute: Optional[UMLAttribute] = None
        self.documentation = ""
        self.is_abstract = False
        self.status = None
        self.composed_of = []
        self.phase = None

        for inherited_stereotype, inherited_package in package.inherited_stereotypes:
            if not hasattr(self, inherited_stereotype):
                setattr(self, inherited_stereotype, inherited_package)

    def __str__(self) -> str:
        return f"{self.name}"

    def get_name(self) -> str:
        if self.alias and settings['use_alias']:
            return f"{self.alias}"
        else:
            return f"{self.name}"

    def get_object_relationships(self) -> list:
        """ Returns all related classes where this class is singular:
                - associations from: one to one, one to many
                - associations to: one to one, many to one
                - generalization
        """
        result = []
        if self.generalization is not None:
            result.append(self.generalization)
        for assoc in self.associations_from:
            if assoc.cardinality in [Cardinality.ONE_TO_MANY, Cardinality.ONE_TO_ONE]:
                result.append(assoc.destination)
        for assoc in self.associations_to:
            if assoc.cardinality in [Cardinality.MANY_TO_ONE, Cardinality.ONE_TO_ONE]:
                result.append(assoc.source)

        return result

    def get_array_relationships(self) -> list:
        """ Returns all related classes where this class is part of an array:
                - associations from: many to one
                - associations to: one to many
        """
        result = []
        for assoc in self.associations_from:
            if assoc.cardinality == Cardinality.MANY_TO_ONE:
                result.append(assoc.destination)
        for assoc in self.associations_to:
            if assoc.cardinality == Cardinality.ONE_TO_MANY:
                result.append(assoc.source)

        return result

    def get_all_attributes(self, abstract_only=True) -> list[UMLAttribute]:
        """ Returns attributes of this class and inherited attrs from generalised classes
        """
        result = self.attributes
        parent = self.generalization
        #if parent is not None:
        #    print(f"    {parent} | {parent.generalization}")
        if parent is not None:
            # print(self)
            # for a in result:
            #     print(f"    {a.name}")
            print(self)
            print(parent)
            attrs = parent.get_all_attributes(abstract_only)

            for a in attrs:
                print(f"    {a.name}")
        return result


class UMLAttribute:
    parent: Union[UMLClass, UMLEnumeration, UMLComponent, UMLInstance]
    name: str
    alias: Optional[str]
    id: Union[int, str]
    is_unique: bool
    stereotype: Optional[str]
    classification: Optional[UMLClass]
    documentation: str
    type: Optional[str]
    value: Optional[str]
    visibility: bool
    is_id: bool
    length: int
    scale: Optional[int]
    precision: Optional[int]
    validations: List[str]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = []

    def __init__(self, parent: Union[UMLClass, UMLEnumeration, UMLComponent, UMLInstance], name: str, id: Union[int, str]):
        self.parent = parent
        self.name = name
        self.alias = None
        self.id = id
        self.is_unique = False
        self.stereotype = None
        self.classification = None
        self.classification_id: Union[None, int, str] = None
        self.documentation = ""
        self.type = None
        self.dest_type: Optional[str] = None
        self.value = None
        self.visibility = True
        self.is_id = False
        self.length = 0
        self.scale = None
        self.precision = None
        self.validations = []

    def __str__(self) -> str:
        return f"{self.name}"

    def get_type(self, translator: Optional[str] = None) -> str:
        if not translator:
            return f"{self.dest_type}"
        if self.type in generation_fields[translator].keys():
            return generation_fields[translator][f"{self.type}"]
        return f"{self.type}"

    def set_type(self, source_type: str):
        # Allow setting of length or scale and precision either: <type> (<precision>,<scale>) or <type> (<length>)
        split: List[str] = source_type.split('(')
        if len(split) > 1:
            source_type = split[0].strip()
            split[1] = split[1].replace(')', '')
            attrs = split[1].split(',')
            if len(attrs) == 1:
                self.length = int(attrs[0])
            elif len(attrs) == 2:
                self.precision = int(attrs[0])
                self.scale = int(attrs[1])
        elif source_type.lower() in ['string', 'str']:
            self.length = 100

        self.type = source_type
        if source_type in generation_fields[settings['generation_type']].keys():
            self.dest_type = generation_fields[settings['generation_type']][source_type]
        else:
            self.dest_type = source_type

    def get_name(self) -> str:
        if self.alias and settings['use_alias']:
            return f"{self.alias}"
        else:
            return f"{self.name}"


class UMLComponent:
    package: UMLPackage
    attributes: List[UMLAttribute]
    stereotype: Optional[str]
    name: str
    id: Union[int, str]
    documentation: str
    status: Optional[UMLStatuses]
    associations_from: List[UMLAssociation]
    associations_to: List[UMLAssociation]
    classification: Optional[UMLClass]

    class Meta:
        id_field = 'id'
        owned_subobjects: List = ['attributes', ]

    def __init__(self, package: UMLPackage, name: str, id: Union[int, str]) -> None:
        self.attributes = []
        self.associations_from = []
        self.associations_to = []
        self.package = package
        self.stereotype = None
        self.name = name
        self.id = id
        self.id_attribute: Optional[UMLAttribute] = None
        self.documentation = ""
        self.status = None
        self.classification_id: Union[None, int, str] = None

    def __str__(self) -> str:
        return f"{self.name}"


def dumps(package: UMLPackage) -> str:
    output = obj_to_dict(package)
    return json.dumps(output)
